@@ -77,14 +77,14 @@ func init() {
 		xlog.Fatalx("creating sherpa prometheus collector", err)
 	}
 
-	adminSherpaHandler, err = sherpa.NewHandler("/admin/api/", moxvar.Version, Admin{}, &adminDoc, &sherpa.HandlerOpts{Collector: collector, AdjustFunctionNames: "none"})
+	adminSherpaHandler, err = sherpa.NewHandler("/api/", moxvar.Version, Admin{}, &adminDoc, &sherpa.HandlerOpts{Collector: collector, AdjustFunctionNames: "none"})
 	if err != nil {
 		xlog.Fatalx("sherpa handler", err)
 	}
 }
 
 // Admin exports web API functions for the admin web interface. All its methods are
-// exported under /admin/api/. Function calls require valid HTTP Authentication
+// exported under api/. Function calls require valid HTTP Authentication
 // credentials of a user.
 type Admin struct{}
 
@@ -132,11 +132,14 @@ func checkAdminAuth(ctx context.Context, passwordfile string, w http.ResponseWri
 	}()
 
 	var err error
+	var remoteIP net.IP
 	addr, err = net.ResolveTCPAddr("tcp", r.RemoteAddr)
 	if err != nil {
 		log.Errorx("parsing remote address", err, mlog.Field("addr", r.RemoteAddr))
+	} else if addr != nil {
+		remoteIP = addr.IP
 	}
-	if addr != nil && !mox.LimiterFailedAuth.Add(addr.IP, start, 1) {
+	if remoteIP != nil && !mox.LimiterFailedAuth.Add(remoteIP, start, 1) {
 		metrics.AuthenticationRatelimitedInc("httpadmin")
 		http.Error(w, "429 - too many auth attempts", http.StatusTooManyRequests)
 		return false
@@ -164,10 +167,12 @@ func checkAdminAuth(ctx context.Context, passwordfile string, w http.ResponseWri
 	}
 	t := strings.SplitN(string(auth), ":", 2)
 	if len(t) != 2 || len(t[1]) < 8 {
+		log.Info("failed authentication attempt", mlog.Field("username", "admin"), mlog.Field("remote", remoteIP))
 		return respondAuthFail()
 	}
 	if err := bcrypt.CompareHashAndPassword([]byte(passwordhash), []byte(t[1])); err != nil {
 		authResult = "badcreds"
+		log.Info("failed authentication attempt", mlog.Field("username", "admin"), mlog.Field("remote", remoteIP))
 		return respondAuthFail()
 	}
 	authCache.lastSuccessHash = passwordhash
@@ -183,7 +188,7 @@ func adminHandle(w http.ResponseWriter, r *http.Request) {
 		return
 	}
 
-	if r.Method == "GET" && r.URL.Path == "/admin/" {
+	if r.Method == "GET" && r.URL.Path == "/" {
 		w.Header().Set("Content-Type", "text/html; charset=utf-8")
 		w.Header().Set("Cache-Control", "no-cache; max-age=0")
 		// We typically return the embedded admin.html, but during development it's handy
@@ -334,8 +339,15 @@ func logPanic(ctx context.Context) {
 }
 
 // return IPs we may be listening on.
-func xlistenIPs(ctx context.Context) []net.IP {
-	ips, err := mox.IPs(ctx)
+func xlistenIPs(ctx context.Context, receiveOnly bool) []net.IP {
+	ips, err := mox.IPs(ctx, receiveOnly)
+	xcheckf(ctx, err, "listing ips")
+	return ips
+}
+
+// return IPs from which we may be sending.
+func xsendingIPs(ctx context.Context) []net.IP {
+	ips, err := mox.IPs(ctx, false)
 	xcheckf(ctx, err, "listing ips")
 	return ips
 }
@@ -361,7 +373,7 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 		panic(&sherpa.Error{Code: "user:notFound", Message: "domain not found"})
 	}
 
-	listenIPs := xlistenIPs(ctx)
+	listenIPs := xlistenIPs(ctx, true)
 	isListenIP := func(ip net.IP) bool {
 		for _, lip := range listenIPs {
 			if ip.Equal(lip) {
@@ -404,6 +416,7 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 			Config: &tls.Config{
 				ServerName: host,
 				MinVersion: tls.VersionTLS12, // ../rfc/8996:31 ../rfc/8997:66
+				RootCAs:    mox.Conf.Static.TLS.CertPool,
 			},
 		}
 		for _, ip := range ips {
@@ -416,6 +429,16 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 		}
 	}
 
+	// If at least one listener with SMTP enabled has specified NATed IPs, we'll skip
+	// some checks related to these IPs.
+	var isNAT bool
+	for _, l := range mox.Conf.Static.Listeners {
+		if l.IPsNATed && l.SMTP.Enabled {
+			isNAT = true
+			break
+		}
+	}
+
 	var wg sync.WaitGroup
 
 	// IPRev
@@ -426,46 +449,75 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 
 		// For each mox.Conf.SpecifiedSMTPListenIPs, and each address for
 		// mox.Conf.HostnameDomain, check if they resolve back to the host name.
-		var ips []net.IP
+		hostIPs := map[dns.Domain][]net.IP{}
 		ips, err := resolver.LookupIP(ctx, "ip", mox.Conf.Static.HostnameDomain.ASCII+".")
 		if err != nil {
 			addf(&r.IPRev.Errors, "Looking up IPs for hostname: %s", err)
 		}
-	nextip:
-		for _, ip := range mox.Conf.Static.SpecifiedSMTPListenIPs {
-			for _, xip := range ips {
-				if ip.Equal(xip) {
-					continue nextip
+		if !isNAT {
+		nextip:
+			for _, ip := range mox.Conf.Static.SpecifiedSMTPListenIPs {
+				for _, xip := range ips {
+					if ip.Equal(xip) {
+						continue nextip
+					}
 				}
+				ips = append(ips, ip)
+			}
+		}
+		hostIPs[mox.Conf.Static.HostnameDomain] = ips
+
+		iplist := func(ips []net.IP) string {
+			var ipstrs []string
+			for _, ip := range ips {
+				ipstrs = append(ipstrs, ip.String())
+			}
+			return strings.Join(ipstrs, ", ")
+		}
+
+		r.IPRev.Hostname = mox.Conf.Static.HostnameDomain
+		r.IPRev.Instructions = []string{
+			fmt.Sprintf("Ensure IPs %s have reverse address %s.", iplist(ips), mox.Conf.Static.HostnameDomain.ASCII),
+		}
+
+		// If we have a socks transport, also check its host and IP.
+		for tname, t := range mox.Conf.Static.Transports {
+			if t.Socks != nil {
+				hostIPs[t.Socks.Hostname] = append(hostIPs[t.Socks.Hostname], t.Socks.IPs...)
+				instr := fmt.Sprintf("For SOCKS transport %s, ensure IPs %s have reverse address %s.", tname, iplist(t.Socks.IPs), t.Socks.Hostname)
+				r.IPRev.Instructions = append(r.IPRev.Instructions, instr)
 			}
-			ips = append(ips, ip)
 		}
 
 		type result struct {
+			Host  dns.Domain
 			IP    string
 			Addrs []string
 			Err   error
 		}
 		results := make(chan result)
-		var ipstrs []string
-		for _, ip := range ips {
-			s := ip.String()
-			go func() {
-				addrs, err := resolver.LookupAddr(ctx, s)
-				results <- result{s, addrs, err}
-			}()
-			ipstrs = append(ipstrs, s)
+		n := 0
+		for host, ips := range hostIPs {
+			for _, ip := range ips {
+				n++
+				s := ip.String()
+				host := host
+				go func() {
+					addrs, err := resolver.LookupAddr(ctx, s)
+					results <- result{host, s, addrs, err}
+				}()
+			}
 		}
 		r.IPRev.IPNames = map[string][]string{}
-		for i := 0; i < len(ips); i++ {
+		for i := 0; i < n; i++ {
 			lr := <-results
-			addrs, ip, err := lr.Addrs, lr.IP, lr.Err
+			host, addrs, ip, err := lr.Host, lr.Addrs, lr.IP, lr.Err
 			if err != nil {
-				addf(&r.IPRev.Errors, "Looking up reverse name for %s: %v", ip, err)
+				addf(&r.IPRev.Errors, "Looking up reverse name for %s of %s: %v", ip, host, err)
 				continue
 			}
 			if len(addrs) != 1 {
-				addf(&r.IPRev.Errors, "Expected exactly 1 name for %s, got %d (%v)", ip, len(addrs), addrs)
+				addf(&r.IPRev.Errors, "Expected exactly 1 name for %s of %s, got %d (%v)", ip, host, len(addrs), addrs)
 			}
 			var match bool
 			for i, a := range addrs {
@@ -475,20 +527,15 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 				if err != nil {
 					addf(&r.IPRev.Errors, "Parsing reverse name %q for %s: %v", a, ip, err)
 				}
-				if ad == mox.Conf.Static.HostnameDomain {
+				if ad == host {
 					match = true
 				}
 			}
 			if !match {
-				addf(&r.IPRev.Errors, "Reverse name(s) %s for ip %s do not match hostname %s, which will cause other mail servers to reject incoming messages from this IP.", strings.Join(addrs, ","), ip, mox.Conf.Static.HostnameDomain)
+				addf(&r.IPRev.Errors, "Reverse name(s) %s for ip %s do not match hostname %s, which will cause other mail servers to reject incoming messages from this IP.", strings.Join(addrs, ","), ip, host)
 			}
 			r.IPRev.IPNames[ip] = addrs
 		}
-
-		r.IPRev.Hostname = mox.Conf.Static.HostnameDomain
-		r.IPRev.Instructions = []string{
-			fmt.Sprintf("Ensure IPs %s have reverse address %s.", strings.Join(ipstrs, ", "), mox.Conf.Static.HostnameDomain.ASCII),
-		}
 	}()
 
 	// MX
@@ -499,7 +546,7 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 
 		mxs, err := resolver.LookupMX(ctx, domain.ASCII+".")
 		if err != nil {
-			addf(&r.MX.Errors, "Looking up MX records for %q: %s", domain, err)
+			addf(&r.MX.Errors, "Looking up MX records for %s: %s", domain, err)
 		}
 		r.MX.Records = make([]MX, len(mxs))
 		for i, mx := range mxs {
@@ -512,9 +559,12 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 		for i, mx := range mxs {
 			ips, ourIPs, notOurIPs, err := lookupIPs(&r.MX.Errors, mx.Host)
 			if err != nil {
-				addf(&r.MX.Errors, "Looking up IP addresses for mx host %q: %s", mx.Host, err)
+				addf(&r.MX.Errors, "Looking up IPs for mx host %q: %s", mx.Host, err)
 			}
 			r.MX.Records[i].IPs = ips
+			if isNAT {
+				continue
+			}
 			if len(ourIPs) == 0 {
 				addf(&r.MX.Errors, "None of the IPs that mx %q points to is ours: %v", mx.Host, notOurIPs)
 			} else if len(notOurIPs) > 0 {
@@ -585,7 +635,11 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 			if !strings.HasPrefix(line, "220 ") {
 				return fmt.Errorf("SMTP STARTTLS response from remote not 220 OK: %q", strings.TrimSuffix(line, "\r\n"))
 			}
-			tlsconn := tls.Client(conn, &tls.Config{ServerName: host})
+			config := &tls.Config{
+				ServerName: host,
+				RootCAs:    mox.Conf.Static.TLS.CertPool,
+			}
+			tlsconn := tls.Client(conn, config)
 			if err := tlsconn.HandshakeContext(cctx); err != nil {
 				return fmt.Errorf("TLS handshake after SMTP STARTTLS: %s", err)
 			}
@@ -623,6 +677,7 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 	}()
 
 	// SPF
+	// todo: add warnings if we have Transports with submission? admin should ensure their IPs are in the SPF record. it may be an IP(net), or an include. that means we cannot easily check for it. and should we first check the transport can be used from this domain (or an account that has this domain?). also see DKIM.
 	wg.Add(1)
 	go func() {
 		defer logPanic(ctx)
@@ -642,38 +697,51 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 			spfr := spf.Record{
 				Version: "spf1",
 			}
+
+			checkSPFIP := func(ip net.IP) {
+				mechanism := "ip4"
+				if ip.To4() == nil {
+					mechanism = "ip6"
+				}
+				spfr.Directives = append(spfr.Directives, spf.Directive{Mechanism: mechanism, IP: ip})
+
+				if record == nil {
+					return
+				}
+
+				args := spf.Args{
+					RemoteIP:          ip,
+					MailFromLocalpart: "postmaster",
+					MailFromDomain:    domain,
+					HelloDomain:       dns.IPDomain{Domain: domain},
+					LocalIP:           net.ParseIP("127.0.0.1"),
+					LocalHostname:     dns.Domain{ASCII: "localhost"},
+				}
+				status, mechanism, expl, err := spf.Evaluate(ctx, record, resolver, args)
+				if err != nil {
+					addf(&r.SPF.Errors, "Evaluating IP %q against %s SPF record: %s", ip, kind, err)
+				} else if status != spf.StatusPass {
+					addf(&r.SPF.Errors, "IP %q does not pass %s SPF evaluation, status not \"pass\" but %q (mechanism %q, explanation %q)", ip, kind, status, mechanism, expl)
+				}
+			}
+
 			for _, l := range mox.Conf.Static.Listeners {
-				if !l.SMTP.Enabled {
+				if !l.SMTP.Enabled || l.IPsNATed {
 					continue
 				}
 				for _, ipstr := range l.IPs {
 					ip := net.ParseIP(ipstr)
-					mechanism := "ip4"
-					if ip.To4() == nil {
-						mechanism = "ip6"
-					}
-					spfr.Directives = append(spfr.Directives, spf.Directive{Mechanism: mechanism, IP: ip})
-
-					if record == nil {
-						continue
-					}
-
-					args := spf.Args{
-						RemoteIP:          ip,
-						MailFromLocalpart: "postmaster",
-						MailFromDomain:    domain,
-						HelloDomain:       dns.IPDomain{Domain: domain},
-						LocalIP:           net.ParseIP("127.0.0.1"),
-						LocalHostname:     dns.Domain{ASCII: "localhost"},
-					}
-					status, mechanism, expl, err := spf.Evaluate(ctx, record, resolver, args)
-					if err != nil {
-						addf(&r.SPF.Errors, "Evaluating IP address %q against %s SPF record: %s", ip, kind, err)
-					} else if status != spf.StatusPass {
-						addf(&r.SPF.Errors, "IP address %q does not pass %s SPF evaluation, status not \"pass\" but %q (mechanism %q, explanation %q)", ip, kind, status, mechanism, expl)
+					checkSPFIP(ip)
+				}
+			}
+			for _, t := range mox.Conf.Static.Transports {
+				if t.Socks != nil {
+					for _, ip := range t.Socks.IPs {
+						checkSPFIP(ip)
 					}
 				}
 			}
+
 			spfr.Directives = append(spfr.Directives, spf.Directive{Qualifier: "-", Mechanism: "all"})
 			return txt, xrecord, spfr
 		}
@@ -697,6 +765,7 @@ func checkDomain(ctx context.Context, resolver dns.Resolver, dialer *net.Dialer,
 	}()
 
 	// DKIM
+	// todo: add warnings if we have Transports with submission? admin should ensure DKIM records exist. we cannot easily check if they actually exist though. and should we first check the transport can be used from this domain (or an account that has this domain?). also see SPF.
 	wg.Add(1)
 	go func() {
 		defer logPanic(ctx)
@@ -1001,7 +1070,7 @@ When enabling MTA-STS, or updating a policy, always update the policy first (thr
 		r.SRVConf.SRVs = map[string][]*net.SRV{}
 		for _, req := range reqs {
 			name := req.name + "_.tcp." + domain.ASCII
-			instr += fmt.Sprintf("\t%s._tcp.%s IN SRV 0 1 %d %s\n", req.name, domain.ASCII+".", req.port, req.host)
+			instr += fmt.Sprintf("\t%s._tcp.%-*s IN SRV 0 1 %d %s\n", req.name, len("_submissions")-len(req.name)+len(domain.ASCII+"."), domain.ASCII+".", req.port, req.host)
 			r.SRVConf.SRVs[req.name] = req.srvs
 			if err != nil {
 				addf(&r.SRVConf.Errors, "Looking up SRV record %q: %s", name, err)
@@ -1030,10 +1099,12 @@ When enabling MTA-STS, or updating a policy, always update the policy first (thr
 		}
 
 		r.Autoconf.IPs = ips
-		if len(ourIPs) == 0 {
-			addf(&r.Autoconf.Errors, "Autoconfig does not point to one of our IP addresses.")
-		} else if len(notOurIPs) > 0 {
-			addf(&r.Autoconf.Errors, "Autoconfig does not point to some IP addresses that are not ours: %v", notOurIPs)
+		if !isNAT {
+			if len(ourIPs) == 0 {
+				addf(&r.Autoconf.Errors, "Autoconfig does not point to one of our IPs.")
+			} else if len(notOurIPs) > 0 {
+				addf(&r.Autoconf.Errors, "Autoconfig points to some IPs that are not ours: %v", notOurIPs)
+			}
 		}
 
 		checkTLS(&r.Autoconf.Errors, "autoconfig."+domain.ASCII, ips, "443")
@@ -1064,10 +1135,12 @@ When enabling MTA-STS, or updating a policy, always update the policy first (thr
 			}
 			match = true
 			r.Autodiscover.Records = append(r.Autodiscover.Records, AutodiscoverSRV{*srv, ips})
-			if len(ourIPs) == 0 {
-				addf(&r.Autodiscover.Errors, "SRV target %q does not point to our IPs.", srv.Target)
-			} else if len(notOurIPs) > 0 {
-				addf(&r.Autodiscover.Errors, "SRV target %q points to some IPs that are not ours: %v", srv.Target, notOurIPs)
+			if !isNAT {
+				if len(ourIPs) == 0 {
+					addf(&r.Autodiscover.Errors, "SRV target %q does not point to our IPs.", srv.Target)
+				} else if len(notOurIPs) > 0 {
+					addf(&r.Autodiscover.Errors, "SRV target %q points to some IPs that are not ours: %v", srv.Target, notOurIPs)
+				}
 			}
 
 			checkTLS(&r.Autodiscover.Errors, strings.TrimSuffix(srv.Target, "."), ips, "443")
@@ -1102,8 +1175,8 @@ func (Admin) Domain(ctx context.Context, domain string) dns.Domain {
 	return d
 }
 
-// DomainLocalparts returns the localparts and accounts configured in domain.
-func (Admin) DomainLocalparts(ctx context.Context, domain string) (localpartAccounts map[smtp.Localpart]string) {
+// DomainLocalparts returns the encoded localparts and accounts configured in domain.
+func (Admin) DomainLocalparts(ctx context.Context, domain string) (localpartAccounts map[string]string) {
 	d, err := dns.ParseDomain(domain)
 	xcheckf(ctx, err, "parsing domain")
 	_, ok := mox.Conf.Domain(d)
@@ -1363,7 +1436,7 @@ func dnsblsStatus(ctx context.Context, resolver dns.Resolver) map[string]map[str
 	}
 
 	r := map[string]map[string]string{}
-	for _, ip := range xlistenIPs(ctx) {
+	for _, ip := range xsendingIPs(ctx) {
 		if ip.IsLoopback() || ip.IsPrivate() {
 			continue
 		}
@@ -1378,7 +1451,7 @@ func dnsblsStatus(ctx context.Context, resolver dns.Resolver) map[string]map[str
 			if expl != "" {
 				result += ": " + expl
 			}
-			r[ipstr][zone.String()] = result
+			r[ipstr][zone.LogString()] = result
 		}
 	}
 	return r
@@ -1457,6 +1530,12 @@ func (Admin) SetPassword(ctx context.Context, accountName, password string) {
 	xcheckf(ctx, err, "setting password")
 }
 
+// SetAccountLimits set new limits on outgoing messages for an account.
+func (Admin) SetAccountLimits(ctx context.Context, accountName string, maxOutgoingMessagesPerDay, maxFirstTimeRecipientsPerDay int) {
+	err := mox.AccountLimitsSave(ctx, accountName, maxOutgoingMessagesPerDay, maxFirstTimeRecipientsPerDay)
+	xcheckf(ctx, err, "saving account limits")
+}
+
 // ClientConfigDomain returns configurations for email clients, IMAP and
 // Submission (SMTP) for the domain.
 func (Admin) ClientConfigDomain(ctx context.Context, domain string) mox.ClientConfig {
@@ -1470,21 +1549,22 @@ func (Admin) ClientConfigDomain(ctx context.Context, domain string) mox.ClientCo
 
 // QueueList returns the messages currently in the outgoing queue.
 func (Admin) QueueList(ctx context.Context) []queue.Msg {
-	l, err := queue.List()
+	l, err := queue.List(ctx)
 	xcheckf(ctx, err, "listing messages in queue")
 	return l
 }
 
 // QueueSize returns the number of messages currently in the outgoing queue.
 func (Admin) QueueSize(ctx context.Context) int {
-	n, err := queue.Count()
+	n, err := queue.Count(ctx)
 	xcheckf(ctx, err, "listing messages in queue")
 	return n
 }
 
-// QueueKick initiates delivery of a message from the queue.
-func (Admin) QueueKick(ctx context.Context, id int64) {
-	n, err := queue.Kick(id, "", "")
+// QueueKick initiates delivery of a message from the queue and sets the transport
+// to use for delivery.
+func (Admin) QueueKick(ctx context.Context, id int64, transport string) {
+	n, err := queue.Kick(ctx, id, "", "", &transport)
 	if err == nil && n == 0 {
 		err = errors.New("message not found")
 	}
@@ -1493,7 +1573,7 @@ func (Admin) QueueKick(ctx context.Context, id int64) {
 
 // QueueDrop removes a message from the queue.
 func (Admin) QueueDrop(ctx context.Context, id int64) {
-	n, err := queue.Drop(id, "", "")
+	n, err := queue.Drop(ctx, id, "", "")
 	if err == nil && n == 0 {
 		err = errors.New("message not found")
 	}
@@ -1597,3 +1677,8 @@ func (Admin) WebserverConfigSave(ctx context.Context, oldConf, newConf Webserver
 	savedConf.WebDomainRedirects = nil
 	return savedConf
 }
+
+// Transports returns the configured transports, for sending email.
+func (Admin) Transports(ctx context.Context) map[string]config.Transport {
+	return mox.Conf.Static.Transports
+}
