@@ -42,7 +42,7 @@ var (
 		},
 		[]string{
 			"handler", // Name from webhandler, can be empty.
-			"proto",   // "http" or "https"
+			"proto",   // "http", "https", "ws", "wss"
 			"method",  // "(unknown)" and otherwise only common verbs
 			"code",
 		},
@@ -58,7 +58,7 @@ var (
 		},
 		[]string{
 			"handler", // Name from webhandler, can be empty.
-			"proto",   // "http" or "https"
+			"proto",   // "http", "https", "ws", "wss"
 			"method",  // "(unknown)" and otherwise only common verbs
 			"code",
 		},
@@ -69,22 +69,37 @@ var (
 
 // http.ResponseWriter that writes access log and tracks metrics at end of response.
 type loggingWriter struct {
-	W     http.ResponseWriter // Calls are forwarded.
-	Start time.Time
-	R     *http.Request
+	W                http.ResponseWriter // Calls are forwarded.
+	Start            time.Time
+	R                *http.Request
+	WebsocketRequest bool // Whether request from was websocket.
 
 	Handler string // Set by router.
 
 	// Set by handlers.
-	StatusCode int
-	Size       int64
-	WriteErr   error
+	StatusCode                   int
+	Size                         int64 // Of data served, for non-websocket responses.
+	Err                          error
+	WebsocketResponse            bool  // If this was a successful websocket connection with backend.
+	SizeFromClient, SizeToClient int64 // Websocket data.
 }
 
 func (w *loggingWriter) Header() http.Header {
 	return w.W.Header()
 }
 
+// protocol, for logging.
+func (w *loggingWriter) proto(websocket bool) string {
+	proto := "http"
+	if websocket {
+		proto = "ws"
+	}
+	if w.R.TLS != nil {
+		proto += "s"
+	}
+	return proto
+}
+
 func (w *loggingWriter) setStatusCode(statusCode int) {
 	if w.StatusCode != 0 {
 		return
@@ -92,11 +107,7 @@ func (w *loggingWriter) setStatusCode(statusCode int) {
 
 	w.StatusCode = statusCode
 	method := metricHTTPMethod(w.R.Method)
-	proto := "http"
-	if w.R.TLS != nil {
-		proto = "https"
-	}
-	metricRequest.WithLabelValues(w.Handler, proto, method, fmt.Sprintf("%d", w.StatusCode)).Observe(float64(time.Since(w.Start)) / float64(time.Second))
+	metricRequest.WithLabelValues(w.Handler, w.proto(w.WebsocketRequest), method, fmt.Sprintf("%d", w.StatusCode)).Observe(float64(time.Since(w.Start)) / float64(time.Second))
 }
 
 func (w *loggingWriter) Write(buf []byte) (int, error) {
@@ -108,8 +119,8 @@ func (w *loggingWriter) Write(buf []byte) (int, error) {
 	if n > 0 {
 		w.Size += int64(n)
 	}
-	if err != nil && w.WriteErr == nil {
-		w.WriteErr = err
+	if err != nil {
+		w.error(err)
 	}
 	return n, err
 }
@@ -136,13 +147,15 @@ func metricHTTPMethod(method string) string {
 	return "(other)"
 }
 
+func (w *loggingWriter) error(err error) {
+	if w.Err == nil {
+		w.Err = err
+	}
+}
+
 func (w *loggingWriter) Done() {
 	method := metricHTTPMethod(w.R.Method)
-	proto := "http"
-	if w.R.TLS != nil {
-		proto = "https"
-	}
-	metricResponse.WithLabelValues(w.Handler, proto, method, fmt.Sprintf("%d", w.StatusCode)).Observe(float64(time.Since(w.Start)) / float64(time.Second))
+	metricResponse.WithLabelValues(w.Handler, w.proto(w.WebsocketResponse), method, fmt.Sprintf("%d", w.StatusCode)).Observe(float64(time.Since(w.Start)) / float64(time.Second))
 
 	tlsinfo := "plain"
 	if w.R.TLS != nil {
@@ -152,33 +165,53 @@ func (w *loggingWriter) Done() {
 			tlsinfo = "(other)"
 		}
 	}
-	xlog.WithContext(w.R.Context()).Debugx("http request", w.WriteErr,
+	err := w.Err
+	if err == nil {
+		err = w.R.Context().Err()
+	}
+	fields := []mlog.Pair{
 		mlog.Field("httpaccess", ""),
 		mlog.Field("handler", w.Handler),
 		mlog.Field("method", method),
 		mlog.Field("url", w.R.URL),
 		mlog.Field("host", w.R.Host),
 		mlog.Field("duration", time.Since(w.Start)),
-		mlog.Field("size", w.Size),
 		mlog.Field("statuscode", w.StatusCode),
 		mlog.Field("proto", strings.ToLower(w.R.Proto)),
 		mlog.Field("remoteaddr", w.R.RemoteAddr),
 		mlog.Field("tlsinfo", tlsinfo),
 		mlog.Field("useragent", w.R.Header.Get("User-Agent")),
 		mlog.Field("referrr", w.R.Header.Get("Referrer")),
-	)
+	}
+	if w.WebsocketRequest {
+		fields = append(fields,
+			mlog.Field("websocketrequest", true),
+		)
+	}
+	if w.WebsocketResponse {
+		fields = append(fields,
+			mlog.Field("websocket", true),
+			mlog.Field("sizetoclient", w.SizeToClient),
+			mlog.Field("sizefromclient", w.SizeFromClient),
+		)
+	} else {
+		fields = append(fields,
+			mlog.Field("size", w.Size),
+		)
+	}
+	xlog.WithContext(w.R.Context()).Debugx("http request", err, fields...)
 }
 
 // Set some http headers that should prevent potential abuse. Better safe than sorry.
-func safeHeaders(fn http.HandlerFunc) http.HandlerFunc {
-	return func(w http.ResponseWriter, r *http.Request) {
+func safeHeaders(fn http.Handler) http.Handler {
+	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		h := w.Header()
 		h.Set("X-Frame-Options", "deny")
 		h.Set("X-Content-Type-Options", "nosniff")
 		h.Set("Content-Security-Policy", "default-src 'self' 'unsafe-inline' data:")
 		h.Set("Referrer-Policy", "same-origin")
-		fn(w, r)
-	}
+		fn.ServeHTTP(w, r)
+	})
 }
 
 // Built-in handlers, e.g. mta-sts and autoconfig.
@@ -186,18 +219,19 @@ type pathHandler struct {
 	Name      string                    // For logging/metrics.
 	HostMatch func(dom dns.Domain) bool // If not nil, called to see if domain of requests matches. Only called if requested host is a valid domain.
 	Path      string                    // Path to register, like on http.ServeMux.
-	Handle    http.HandlerFunc
+	Handler   http.Handler
 }
 type serve struct {
-	Kinds        []string // Type of handler and protocol (http/https).
+	Kinds        []string // Type of handler and protocol (e.g. acme-tls-alpn-01, account-http, admin-https).
 	TLSConfig    *tls.Config
 	PathHandlers []pathHandler // Sorted, longest first.
 	Webserver    bool          // Whether serving WebHandler. PathHandlers are always evaluated before WebHandlers.
 }
 
-// HandleFunc registers a named handler for a path and optional host. If path ends with a slash, it
-// is used as prefix match, otherwise a full path match is required. If hostOpt is set, only requests to those host are handled by this handler.
-func (s *serve) HandleFunc(name string, hostMatch func(dns.Domain) bool, path string, fn http.HandlerFunc) {
+// Handle registers a named handler for a path and optional host. If path ends with
+// a slash, it is used as prefix match, otherwise a full path match is required. If
+// hostOpt is set, only requests to those host are handled by this handler.
+func (s *serve) Handle(name string, hostMatch func(dns.Domain) bool, path string, fn http.Handler) {
 	s.PathHandlers = append(s.PathHandlers, pathHandler{name, hostMatch, path, fn})
 }
 
@@ -278,7 +312,7 @@ func (s *serve) ServeHTTP(xw http.ResponseWriter, r *http.Request) {
 		}
 		if r.URL.Path == h.Path || strings.HasSuffix(h.Path, "/") && strings.HasPrefix(r.URL.Path, h.Path) {
 			nw.Handler = h.Name
-			h.Handle(nw, r)
+			h.Handler.ServeHTTP(nw, r)
 			return
 		}
 	}
@@ -294,6 +328,16 @@ func (s *serve) ServeHTTP(xw http.ResponseWriter, r *http.Request) {
 // Listen binds to sockets for HTTP listeners, including those required for ACME to
 // generate TLS certificates. It stores the listeners so Serve can start serving them.
 func Listen() {
+	redirectToTrailingSlash := func(srv *serve, name, path string) {
+		// Helpfully redirect user to version with ending slash.
+		if path != "/" && strings.HasSuffix(path, "/") {
+			handler := safeHeaders(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+				http.Redirect(w, r, path, http.StatusSeeOther)
+			}))
+			srv.Handle(name, nil, path[:len(path)-1], handler)
+		}
+	}
+
 	for name, l := range mox.Conf.Static.Listeners {
 		portServe := map[int]*serve{}
 
@@ -319,41 +363,59 @@ func Listen() {
 
 		if l.TLS != nil && l.TLS.ACME != "" && (l.SMTP.Enabled && !l.SMTP.NoSTARTTLS || l.Submissions.Enabled || l.IMAPS.Enabled) {
 			port := config.Port(mox.Conf.Static.ACME[l.TLS.ACME].Port, 443)
-			ensureServe(true, port, "acme-tls-alpn01")
+			ensureServe(true, port, "acme-tls-alpn-01")
 		}
 
 		if l.AccountHTTP.Enabled {
 			port := config.Port(l.AccountHTTP.Port, 80)
-			srv := ensureServe(false, port, "account-http")
-			srv.HandleFunc("account", nil, "/", safeHeaders(accountHandle))
+			path := "/"
+			if l.AccountHTTP.Path != "" {
+				path = l.AccountHTTP.Path
+			}
+			srv := ensureServe(false, port, "account-http at "+path)
+			handler := safeHeaders(http.StripPrefix(path[:len(path)-1], http.HandlerFunc(accountHandle)))
+			srv.Handle("account", nil, path, handler)
+			redirectToTrailingSlash(srv, "account", path)
 		}
 		if l.AccountHTTPS.Enabled {
 			port := config.Port(l.AccountHTTPS.Port, 443)
-			srv := ensureServe(true, port, "account-https")
-			srv.HandleFunc("account", nil, "/", safeHeaders(accountHandle))
+			path := "/"
+			if l.AccountHTTPS.Path != "" {
+				path = l.AccountHTTPS.Path
+			}
+			srv := ensureServe(true, port, "account-https at "+path)
+			handler := safeHeaders(http.StripPrefix(path[:len(path)-1], http.HandlerFunc(accountHandle)))
+			srv.Handle("account", nil, path, handler)
+			redirectToTrailingSlash(srv, "account", path)
 		}
 
 		if l.AdminHTTP.Enabled {
 			port := config.Port(l.AdminHTTP.Port, 80)
-			srv := ensureServe(false, port, "admin-http")
-			if !l.AccountHTTP.Enabled {
-				srv.HandleFunc("admin", nil, "/", safeHeaders(adminIndex))
+			path := "/admin/"
+			if l.AdminHTTP.Path != "" {
+				path = l.AdminHTTP.Path
 			}
-			srv.HandleFunc("admin", nil, "/admin/", safeHeaders(adminHandle))
+			srv := ensureServe(false, port, "admin-http at "+path)
+			handler := safeHeaders(http.StripPrefix(path[:len(path)-1], http.HandlerFunc(adminHandle)))
+			srv.Handle("admin", nil, path, handler)
+			redirectToTrailingSlash(srv, "admin", path)
 		}
 		if l.AdminHTTPS.Enabled {
 			port := config.Port(l.AdminHTTPS.Port, 443)
-			srv := ensureServe(true, port, "admin-https")
-			if !l.AccountHTTPS.Enabled {
-				srv.HandleFunc("admin", nil, "/", safeHeaders(adminIndex))
+			path := "/admin/"
+			if l.AdminHTTPS.Path != "" {
+				path = l.AdminHTTPS.Path
 			}
-			srv.HandleFunc("admin", nil, "/admin/", safeHeaders(adminHandle))
+			srv := ensureServe(true, port, "admin-https at "+path)
+			handler := safeHeaders(http.StripPrefix(path[:len(path)-1], http.HandlerFunc(adminHandle)))
+			srv.Handle("admin", nil, path, handler)
+			redirectToTrailingSlash(srv, "admin", path)
 		}
 		if l.MetricsHTTP.Enabled {
 			port := config.Port(l.MetricsHTTP.Port, 8010)
 			srv := ensureServe(false, port, "metrics-http")
-			srv.HandleFunc("metrics", nil, "/metrics", safeHeaders(promhttp.Handler().ServeHTTP))
-			srv.HandleFunc("metrics", nil, "/", safeHeaders(func(w http.ResponseWriter, r *http.Request) {
+			srv.Handle("metrics", nil, "/metrics", safeHeaders(promhttp.Handler()))
+			srv.Handle("metrics", nil, "/", safeHeaders(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 				if r.URL.Path != "/" {
 					http.NotFound(w, r)
 					return
@@ -363,7 +425,7 @@ func Listen() {
 				}
 				w.Header().Set("Content-Type", "text/html")
 				fmt.Fprint(w, `<html><body>see <a href="/metrics">/metrics</a></body></html>`)
-			}))
+			})))
 		}
 		if l.AutoconfigHTTPS.Enabled {
 			port := config.Port(l.AutoconfigHTTPS.Port, 443)
@@ -372,17 +434,17 @@ func Listen() {
 				// todo: may want to check this against the configured domains, could in theory be just a webserver.
 				return strings.HasPrefix(dom.ASCII, "autoconfig.")
 			}
-			srv.HandleFunc("autoconfig", autoconfigMatch, "/mail/config-v1.1.xml", safeHeaders(autoconfHandle))
-			srv.HandleFunc("autodiscover", autoconfigMatch, "/autodiscover/autodiscover.xml", safeHeaders(autodiscoverHandle))
+			srv.Handle("autoconfig", autoconfigMatch, "/mail/config-v1.1.xml", safeHeaders(http.HandlerFunc(autoconfHandle)))
+			srv.Handle("autodiscover", autoconfigMatch, "/autodiscover/autodiscover.xml", safeHeaders(http.HandlerFunc(autodiscoverHandle)))
 		}
 		if l.MTASTSHTTPS.Enabled {
 			port := config.Port(l.MTASTSHTTPS.Port, 443)
-			srv := ensureServe(!l.AutoconfigHTTPS.NonTLS, port, "mtasts-https")
+			srv := ensureServe(!l.MTASTSHTTPS.NonTLS, port, "mtasts-https")
 			mtastsMatch := func(dom dns.Domain) bool {
 				// todo: may want to check this against the configured domains, could in theory be just a webserver.
 				return strings.HasPrefix(dom.ASCII, "mta-sts.")
 			}
-			srv.HandleFunc("mtasts", mtastsMatch, "/.well-known/mta-sts.txt", safeHeaders(mtastsPolicyHandle))
+			srv.Handle("mtasts", mtastsMatch, "/.well-known/mta-sts.txt", safeHeaders(http.HandlerFunc(mtastsPolicyHandle)))
 		}
 		if l.PprofHTTP.Enabled {
 			// Importing net/http/pprof registers handlers on the default serve mux.
@@ -392,7 +454,7 @@ func Listen() {
 			}
 			srv := &serve{[]string{"pprof-http"}, nil, nil, false}
 			portServe[port] = srv
-			srv.HandleFunc("pprof", nil, "/", http.DefaultServeMux.ServeHTTP)
+			srv.Handle("pprof", nil, "/", http.DefaultServeMux)
 		}
 		if l.WebserverHTTP.Enabled {
 			port := config.Port(l.WebserverHTTP.Port, 80)
@@ -405,16 +467,16 @@ func Listen() {
 			srv.Webserver = true
 		}
 
-		// We'll explicitly ensure these TLS certs exist (e.g. are created with ACME)
-		// immediately after startup. We only do so for our explicit listener hostnames,
-		// not for mta-sts DNS records, it can be requested on demand (perhaps never). We
-		// do request autoconfig, otherwise clients may run into their timeouts waiting for
-		// the certificate to be given during the first https connection.
-		ensureManagerHosts = map[*autotls.Manager]map[dns.Domain]struct{}{}
-
 		if l.TLS != nil && l.TLS.ACME != "" {
 			m := mox.Conf.Static.ACME[l.TLS.ACME].Manager
 
+			// If we are listening on port 80 for plain http, also register acme http-01
+			// validation handler.
+			if srv, ok := portServe[80]; ok && srv.TLSConfig == nil {
+				srv.Kinds = append(srv.Kinds, "acme-http-01")
+				srv.Handle("acme-http-01", nil, "/.well-known/acme-challenge/", m.Manager.HTTPHandler(nil))
+			}
+
 			hosts := map[dns.Domain]struct{}{
 				mox.Conf.Static.HostnameDomain: {},
 			}
@@ -452,22 +514,15 @@ func Listen() {
 	}
 }
 
-// Only used when the account page is not active on the same listener.
-func adminIndex(w http.ResponseWriter, r *http.Request) {
-	if r.URL.Path != "/" {
-		http.NotFound(w, r)
-		return
-	}
-	if r.Method != "GET" {
-		http.Error(w, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)
-		return
-	}
-	http.Redirect(w, r, "/admin/", http.StatusSeeOther)
-}
-
 // functions to be launched in goroutine that will serve on a listener.
 var servers []func()
-var ensureManagerHosts map[*autotls.Manager]map[dns.Domain]struct{}
+
+// We'll explicitly ensure these TLS certs exist (e.g. are created with ACME)
+// immediately after startup. We only do so for our explicit listener hostnames,
+// not for mta-sts DNS records, it can be requested on demand (perhaps never). We
+// do request autoconfig, otherwise clients may run into their timeouts waiting for
+// the certificate to be given during the first https connection.
+var ensureManagerHosts = map[*autotls.Manager]map[dns.Domain]struct{}{}
 
 // listen prepares a listener, and adds it to "servers", to be launched (if not running as root) through Serve.
 func listen1(ip string, port int, tlsConfig *tls.Config, name string, kinds []string, handler http.Handler) {
@@ -498,9 +553,11 @@ func listen1(ip string, port int, tlsConfig *tls.Config, name string, kinds []st
 	}
 
 	server := &http.Server{
-		Handler:   handler,
-		TLSConfig: tlsConfig,
-		ErrorLog:  golog.New(mlog.ErrWriter(xlog.Fields(mlog.Field("pkg", "net/http")), mlog.LevelInfo, protocol+" error"), "", 0),
+		Handler:           handler,
+		TLSConfig:         tlsConfig,
+		ReadHeaderTimeout: 30 * time.Second,
+		IdleTimeout:       65 * time.Second, // Chrome closes connections after 60 seconds, firefox after 115 seconds.
+		ErrorLog:          golog.New(mlog.ErrWriter(xlog.Fields(mlog.Field("pkg", "net/http")), mlog.LevelInfo, protocol+" error"), "", 0),
 	}
 	serve := func() {
 		err := server.Serve(ln)
