@@ -5,6 +5,7 @@ import (
 	"encoding/base64"
 	"errors"
 	"fmt"
+	"io"
 	"net"
 	"os"
 	"testing"
@@ -58,8 +59,8 @@ func FuzzServer(f *testing.F) {
 	}
 
 	mox.Context = context.Background()
-	mox.ConfigStaticPath = "../testdata/imap/mox.conf"
-	mox.MustLoadConfig(false)
+	mox.ConfigStaticPath = "../testdata/imapserverfuzz/mox.conf"
+	mox.MustLoadConfig(true, false)
 	dataDir := mox.ConfigDirPath(mox.Conf.Static.DataDir)
 	os.RemoveAll(dataDir)
 	acc, err := store.OpenAccount("mjl")
@@ -81,7 +82,7 @@ func FuzzServer(f *testing.F) {
 
 	var fl *os.File
 	if false {
-		fl, err = os.Create("fuzz.log")
+		fl, err = os.OpenFile("fuzz.log", os.O_CREATE|os.O_APPEND|os.O_WRONLY, 0600)
 		if err != nil {
 			f.Fatalf("fuzz log")
 		}
@@ -95,6 +96,7 @@ func FuzzServer(f *testing.F) {
 
 	f.Fuzz(func(t *testing.T, s string) {
 		run := func(cmds []string) {
+			limitersInit() // Reset rate limiters.
 			serverConn, clientConn := net.Pipe()
 			defer serverConn.Close()
 
@@ -106,7 +108,7 @@ func FuzzServer(f *testing.F) {
 						return
 					}
 					err, ok := x.(error)
-					if !ok || !errors.Is(err, os.ErrDeadlineExceeded) {
+					if !ok || (!errors.Is(err, os.ErrDeadlineExceeded) && !errors.Is(err, io.EOF)) {
 						panic(x)
 					}
 				}()
@@ -131,6 +133,8 @@ func FuzzServer(f *testing.F) {
 			cid++
 		}
 
+		// Each command brings the connection state one step further. We try the fuzzing
+		// input for each state.
 		run([]string{})
 		run([]string{"login mjl@mox.example testtest"})
 		run([]string{"login mjl@mox.example testtest", "select inbox"})
