@@ -1,6 +1,7 @@
 package imapserver
 
 import (
+	"errors"
 	"fmt"
 	"net/textproto"
 	"strconv"
@@ -11,12 +12,13 @@ import (
 )
 
 var (
-	listWildcards = "%*"
-	char          = charRange('\x01', '\x7f')
-	ctl           = charRange('\x01', '\x19')
-	atomChar      = charRemove(char, "(){ "+listWildcards+ctl)
-	respSpecials  = atomChar + "]"
-	astringChar   = atomChar + respSpecials
+	listWildcards  = "%*"
+	char           = charRange('\x01', '\x7f')
+	ctl            = charRange('\x01', '\x19')
+	quotedSpecials = `"\`
+	respSpecials   = "]"
+	atomChar       = charRemove(char, "(){ "+ctl+listWildcards+quotedSpecials+respSpecials)
+	astringChar    = atomChar + respSpecials
 )
 
 func charRange(first, last rune) string {
@@ -74,11 +76,20 @@ func newParser(s string, conn *conn) *parser {
 }
 
 func (p *parser) xerrorf(format string, args ...any) {
-	var context string
+	var err error
+	errmsg := fmt.Sprintf(format, args...)
+	remaining := fmt.Sprintf("remaining %q", p.orig[p.o:])
 	if len(p.contexts) > 0 {
-		context = strings.Join(p.contexts, ",")
+		remaining += ", context " + strings.Join(p.contexts, ",")
 	}
-	panic(syntaxError{"", "", fmt.Errorf("%s (%sremaining data %q)", fmt.Sprintf(format, args...), context, p.orig[p.o:])})
+	remaining = " (" + remaining + ")"
+	if p.conn.account != nil {
+		errmsg += remaining
+		err = errors.New(errmsg)
+	} else {
+		err = errors.New(errmsg + remaining)
+	}
+	panic(syntaxError{"", "", errmsg, err})
 }
 
 func (p *parser) context(s string) func() {
@@ -112,7 +123,7 @@ func (p *parser) take(s string) bool {
 
 func (p *parser) xtake(s string) {
 	if !p.take(s) {
-		p.xerrorf("expected %q", s)
+		p.xerrorf("expected %s", s)
 	}
 }
 
@@ -191,9 +202,10 @@ func (p *parser) xspace() {
 func (p *parser) digits() string {
 	var n int
 	for _, c := range p.upper[p.o:] {
-		if c >= '0' && c <= '9' {
-			n++
+		if c < '0' || c > '9' {
+			break
 		}
+		n++
 	}
 	if n == 0 {
 		return ""
@@ -366,8 +378,18 @@ func (p *parser) remainder() string {
 	return p.orig[p.o:]
 }
 
+// ../rfc/9051:6565
 func (p *parser) xflag() string {
-	return p.xtakelist(`\`, "$") + p.xatom()
+	w, _ := p.takelist(`\`, "$")
+	s := w + p.xatom()
+	if s[0] == '\\' {
+		switch strings.ToLower(s) {
+		case `\answered`, `\flagged`, `\deleted`, `\seen`, `\draft`:
+		default:
+			p.xerrorf("unknown system flag %s", s)
+		}
+	}
+	return s
 }
 
 func (p *parser) xflagList() (l []string) {
@@ -724,7 +746,8 @@ func (p *parser) xliteralSize(maxSize int64, lit8 bool) (size int64, sync bool)
 	if maxSize > 0 && size > maxSize {
 		// ../rfc/7888:249
 		line := fmt.Sprintf("* BYE [ALERT] Max literal size %d is larger than allowed %d in this context", size, maxSize)
-		panic(syntaxError{line, "TOOBIG", fmt.Errorf("literal too big")})
+		err := errors.New("literal too big")
+		panic(syntaxError{line, "TOOBIG", err.Error(), err})
 	}
 
 	sync = !p.take("+")
