@@ -26,6 +26,7 @@ func init() {
 
 	// Don't slow down tests.
 	badClientDelay = 0
+	authFailDelay = 0
 }
 
 func tocrlf(s string) string {
@@ -190,6 +191,11 @@ func (tc *testconn) xcodeArg(v any) {
 }
 
 func (tc *testconn) xuntagged(exps ...any) {
+	tc.t.Helper()
+	tc.xuntaggedCheck(true, exps...)
+}
+
+func (tc *testconn) xuntaggedCheck(all bool, exps ...any) {
 	tc.t.Helper()
 	last := append([]imapclient.Untagged{}, tc.lastUntagged...)
 next:
@@ -211,7 +217,7 @@ next:
 		}
 		tc.t.Fatalf("did not find untagged response %#v %T (%d) in %v%s", exp, exp, ei, tc.lastUntagged, next)
 	}
-	if len(last) > 0 {
+	if len(last) > 0 && all {
 		tc.t.Fatalf("leftover untagged responses %v", last)
 	}
 }
@@ -310,7 +316,7 @@ func startArgs(t *testing.T, first, isTLS, allowLoginWithoutTLS bool) *testconn
 	}
 	mox.Context = context.Background()
 	mox.ConfigStaticPath = "../testdata/imap/mox.conf"
-	mox.MustLoadConfig(false)
+	mox.MustLoadConfig(true, false)
 	acc, err := store.OpenAccount("mjl")
 	tcheck(t, err, "open account")
 	if first {
@@ -386,8 +392,12 @@ func TestLogin(t *testing.T) {
 	tc.close()
 
 	tc = start(t)
-	defer tc.close()
 	tc.transactf("ok", `login "mjl@mox.example" "testtest"`)
+	tc.close()
+
+	tc = start(t)
+	tc.transactf("ok", `login "\"\"@mox.example" "testtest"`)
+	defer tc.close()
 
 	tc.transactf("bad", "logout badarg")
 	tc.transactf("ok", "logout")
@@ -397,8 +407,6 @@ func TestLogin(t *testing.T) {
 func TestState(t *testing.T) {
 	tc := start(t)
 
-	tc.transactf("bad", "boguscommand")
-
 	notAuthenticated := []string{"starttls", "authenticate", "login"}
 	authenticatedOrSelected := []string{"enable", "select", "examine", "create", "delete", "rename", "subscribe", "unsubscribe", "list", "namespace", "status", "append", "idle", "lsub"}
 	selected := []string{"close", "unselect", "expunge", "search", "fetch", "store", "copy", "move", "uid expunge"}
@@ -421,6 +429,21 @@ func TestState(t *testing.T) {
 	for _, cmd := range append(append([]string{}, notAuthenticated...), selected...) {
 		tc.transactf("no", "%s", cmd)
 	}
+
+	tc.transactf("bad", "boguscommand")
+}
+
+func TestNonIMAP(t *testing.T) {
+	tc := start(t)
+	defer tc.close()
+
+	// imap greeting has already been read, we sidestep the imapclient.
+	_, err := fmt.Fprintf(tc.conn, "bogus\r\n")
+	tc.check(err, "write bogus command")
+	tc.readprefixline("* BYE ")
+	if _, err := tc.conn.Read(make([]byte, 1)); err == nil {
+		t.Fatalf("connection not closed after initial bad command")
+	}
 }
 
 func TestLiterals(t *testing.T) {
@@ -507,7 +530,7 @@ func TestScenario(t *testing.T) {
 	tc.transactf("ok", `store 1 flags.silent (\seen \answered)`)
 	tc.transactf("ok", `store 1 -flags.silent (\answered)`)
 	tc.transactf("ok", `store 1 +flags.silent (\answered)`)
-	tc.transactf("no", `store 1 flags (\badflag)`)
+	tc.transactf("bad", `store 1 flags (\badflag)`)
 	tc.transactf("ok", "noop")
 
 	tc.transactf("ok", "copy 1 Trash")
