@@ -1,12 +1,10 @@
 package main
 
 import (
-	"bufio"
 	"bytes"
 	"context"
 	"crypto/ed25519"
 	"crypto/rsa"
-	"crypto/tls"
 	"crypto/x509"
 	"encoding/base64"
 	"encoding/json"
@@ -16,9 +14,9 @@ import (
 	"io"
 	"log"
 	"net"
-	"net/mail"
 	"os"
 	"path/filepath"
+	"strconv"
 	"strings"
 	"time"
 
@@ -42,7 +40,6 @@ import (
 	"github.com/mjl-/mox/moxvar"
 	"github.com/mjl-/mox/mtasts"
 	"github.com/mjl-/mox/smtp"
-	"github.com/mjl-/mox/smtpclient"
 	"github.com/mjl-/mox/spf"
 	"github.com/mjl-/mox/store"
 	"github.com/mjl-/mox/tlsrpt"
@@ -90,7 +87,10 @@ var commands = []struct {
 	{"import mbox", cmdImportMbox},
 	{"export maildir", cmdExportMaildir},
 	{"export mbox", cmdExportMbox},
+	{"localserve", cmdLocalserve},
 	{"help", cmdHelp},
+	{"backup", cmdBackup},
+	{"verifydata", cmdVerifydata},
 
 	{"config test", cmdConfigTest},
 	{"config dnscheck", cmdConfigDNSCheck},
@@ -116,6 +116,7 @@ var commands = []struct {
 	{"dkim lookup", cmdDKIMLookup},
 	{"dkim txt", cmdDKIMTXT},
 	{"dkim verify", cmdDKIMVerify},
+	{"dkim sign", cmdDKIMSign},
 	{"dmarc lookup", cmdDMARCLookup},
 	{"dmarc parsereportmsg", cmdDMARCParsereportmsg},
 	{"dmarc verify", cmdDMARCVerify},
@@ -139,14 +140,20 @@ var commands = []struct {
 	{"junk test", cmdJunkTest},
 	{"junk train", cmdJunkTrain},
 	{"bumpuidvalidity", cmdBumpUIDValidity},
+	{"reassignuids", cmdReassignUIDs},
+	{"fixuidmeta", cmdFixUIDMeta},
 	{"dmarcdb addreport", cmdDMARCDBAddReport},
 	{"ensureparsed", cmdEnsureParsed},
+	{"message parse", cmdMessageParse},
 	{"tlsrptdb addreport", cmdTLSRPTDBAddReport},
 	{"updates addsigned", cmdUpdatesAddSigned},
 	{"updates genkey", cmdUpdatesGenkey},
 	{"updates pubkey", cmdUpdatesPubkey},
 	{"updates serve", cmdUpdatesServe},
 	{"updates verify", cmdUpdatesVerify},
+	{"gentestdata", cmdGentestdata},
+	{"ximport maildir", cmdXImportMaildir},
+	{"ximport mbox", cmdXImportMbox},
 }
 
 var cmds []cmd
@@ -313,18 +320,18 @@ Used to generate documentation.
 
 		fmt.Fprintf(os.Stderr, "# mox %s\n\n", strings.Join(c.words, " "))
 		if c.help != "" {
-			fmt.Println(c.help + "\n")
+			fmt.Fprintln(os.Stderr, c.help+"\n")
 		}
 		s := c.makeUsage()
 		s = "\t" + strings.ReplaceAll(s, "\n", "\n\t")
-		fmt.Println(s)
+		fmt.Fprintln(os.Stderr, s)
 	}
 }
 
 func usage(l []cmd, unlisted bool) {
 	var lines []string
 	if !unlisted {
-		lines = append(lines, "mox [-config config/mox.conf] ...")
+		lines = append(lines, "mox [-config config/mox.conf] [-pedantic] ...")
 	}
 	for _, c := range l {
 		c.gather()
@@ -350,15 +357,21 @@ func usage(l []cmd, unlisted bool) {
 }
 
 var loglevel string
+var pedantic bool
 
 // subcommands that are not "serve" should use this function to load the config, it
 // restores any loglevel specified on the command-line, instead of using the
-// loglevels from the config file.
+// loglevels from the config file and it does not load files like TLS keys/certs.
 func mustLoadConfig() {
-	mox.MustLoadConfig(false)
-	if level, ok := mlog.Levels[loglevel]; ok && loglevel != "" {
+	mox.MustLoadConfig(false, false)
+	if level, ok := mlog.Levels[loglevel]; loglevel != "" && ok {
 		mox.Conf.Log[""] = level
 		mlog.SetConfig(mox.Conf.Log)
+	} else if loglevel != "" && !ok {
+		log.Fatal("unknown loglevel", mlog.Field("loglevel", loglevel))
+	}
+	if pedantic {
+		moxvar.Pedantic = true
 	}
 }
 
@@ -378,6 +391,11 @@ func main() {
 
 	flag.StringVar(&mox.ConfigStaticPath, "config", envString("MOXCONF", "config/mox.conf"), "configuration file, other config files are looked up in the same directory, defaults to $MOXCONF with a fallback to mox.conf")
 	flag.StringVar(&loglevel, "loglevel", "", "if non-empty, this log level is set early in startup")
+	flag.BoolVar(&pedantic, "pedantic", false, "protocol violations result in errors instead of accepting/working around them")
+
+	var cpuprofile, memprofile string
+	flag.StringVar(&cpuprofile, "cpuprof", "", "store cpu profile to file")
+	flag.StringVar(&memprofile, "memprof", "", "store mem profile to file")
 
 	flag.Usage = func() { usage(cmds, false) }
 	flag.Parse()
@@ -386,6 +404,12 @@ func main() {
 		usage(cmds, false)
 	}
 
+	defer profile(cpuprofile, memprofile)()
+
+	if pedantic {
+		moxvar.Pedantic = true
+	}
+
 	mox.ConfigDynamicPath = filepath.Join(filepath.Dir(mox.ConfigStaticPath), "domains.conf")
 	if level, ok := mlog.Levels[loglevel]; ok && loglevel != "" {
 		mox.Conf.Log[""] = level
@@ -568,17 +592,20 @@ must be set if and only if account does not yet exist.
 
 	d := xparseDomain(args[0], "domain")
 	mustLoadConfig()
+	var localpart string
+	if len(args) == 3 {
+		localpart = args[2]
+	}
+	ctlcmdConfigDomainAdd(xctl(), d, args[1], localpart)
+}
 
-	if len(args) == 2 {
-		args = append(args, "")
-	}
-	ctl := xctl()
+func ctlcmdConfigDomainAdd(ctl *ctl, domain dns.Domain, account, localpart string) {
 	ctl.xwrite("domainadd")
-	for _, s := range args {
-		ctl.xwrite(s)
-	}
+	ctl.xwrite(domain.Name())
+	ctl.xwrite(account)
+	ctl.xwrite(localpart)
 	ctl.xreadok()
-	fmt.Printf("domain added, remember to add dns records, see:\n\nmox config dnsrecords %s\nmox config dnscheck %s\n", d.Name(), d.Name())
+	fmt.Printf("domain added, remember to add dns records, see:\n\nmox config dnsrecords %s\nmox config dnscheck %s\n", domain.Name(), domain.Name())
 }
 
 func cmdConfigDomainRemove(c *cmd) {
@@ -595,9 +622,12 @@ rejected.
 
 	d := xparseDomain(args[0], "domain")
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdConfigDomainRemove(xctl(), d)
+}
+
+func ctlcmdConfigDomainRemove(ctl *ctl, d dns.Domain) {
 	ctl.xwrite("domainrm")
-	ctl.xwrite(args[0])
+	ctl.xwrite(d.Name())
 	ctl.xreadok()
 	fmt.Printf("domain removed, remember to remove dns records for %s\n", d)
 }
@@ -615,13 +645,15 @@ explicitly, see the setaccountpassword command.
 	}
 
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdConfigAccountAdd(xctl(), args[0], args[1])
+}
+
+func ctlcmdConfigAccountAdd(ctl *ctl, account, address string) {
 	ctl.xwrite("accountadd")
-	for _, s := range args {
-		ctl.xwrite(s)
-	}
+	ctl.xwrite(account)
+	ctl.xwrite(address)
 	ctl.xreadok()
-	fmt.Printf("account added, set a password with \"mox setaccountpassword %s\"\n", args[1])
+	fmt.Printf("account added, set a password with \"mox setaccountpassword %s\"\n", address)
 }
 
 func cmdConfigAccountRemove(c *cmd) {
@@ -637,27 +669,36 @@ these addresses will be rejected.
 	}
 
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdConfigAccountRemove(xctl(), args[0])
+}
+
+func ctlcmdConfigAccountRemove(ctl *ctl, account string) {
 	ctl.xwrite("accountrm")
-	ctl.xwrite(args[0])
+	ctl.xwrite(account)
 	ctl.xreadok()
 	fmt.Println("account removed")
 }
 
 func cmdConfigAddressAdd(c *cmd) {
 	c.params = "address account"
-	c.help = "Adds an address to an account and reloads the configuration."
+	c.help = `Adds an address to an account and reloads the configuration.
+
+If address starts with a @ (i.e. a missing localpart), this is a catchall
+address for the domain.
+`
 	args := c.Parse()
 	if len(args) != 2 {
 		c.Usage()
 	}
 
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdConfigAddressAdd(xctl(), args[0], args[1])
+}
+
+func ctlcmdConfigAddressAdd(ctl *ctl, address, account string) {
 	ctl.xwrite("addressadd")
-	for _, s := range args {
-		ctl.xwrite(s)
-	}
+	ctl.xwrite(address)
+	ctl.xwrite(account)
 	ctl.xreadok()
 	fmt.Println("address added")
 }
@@ -666,7 +707,7 @@ func cmdConfigAddressRemove(c *cmd) {
 	c.params = "address"
 	c.help = `Remove an address and reload the configuration.
 
-Incoming email for this address will be rejected.
+Incoming email for this address will be rejected after removing an address.
 `
 	args := c.Parse()
 	if len(args) != 1 {
@@ -674,9 +715,12 @@ Incoming email for this address will be rejected.
 	}
 
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdConfigAddressRemove(xctl(), args[0])
+}
+
+func ctlcmdConfigAddressRemove(ctl *ctl, address string) {
 	ctl.xwrite("addressrm")
-	ctl.xwrite(args[0])
+	ctl.xwrite(address)
 	ctl.xreadok()
 	fmt.Println("address removed")
 }
@@ -775,6 +819,20 @@ WebDomainRedirects:
 
 # Each request is matched against these handlers until one matches and serves it.
 WebHandlers:
+	-
+		# Redirect all plain http requests to https, leaving path, query strings, etc
+		# intact. When the request is already to https, the destination URL would have the
+		# same scheme, host and path, causing this redirect handler to not match the
+		# request (and not cause a redirect loop) and the webserver to serve the request
+		# with a later handler.
+		LogName: redirhttps
+		Domain: www.mox.example
+		PathRegexp: ^/
+		# Could leave DontRedirectPlainHTTP at false if it wasn't for this being an
+		# example for doing this redirect.
+		DontRedirectPlainHTTP: true
+		WebRedirect:
+			BaseURL: https://www.mox.example
 	-
 		# The name of the handler, used in logging and metrics.
 		LogName: staticmjl
@@ -889,21 +947,26 @@ Valid labels: error, info, debug, trace, traceauth, tracedata.
 	mustLoadConfig()
 
 	if len(args) == 0 {
-		ctl := xctl()
-		ctl.xwrite("loglevels")
-		ctl.xreadok()
-		ctl.xstreamto(os.Stdout)
-		return
-	}
-
-	ctl := xctl()
-	ctl.xwrite("setloglevels")
-	if len(args) == 2 {
-		ctl.xwrite(args[1])
+		ctlcmdLoglevels(xctl())
 	} else {
-		ctl.xwrite("")
+		var pkg string
+		if len(args) == 2 {
+			pkg = args[1]
+		}
+		ctlcmdSetLoglevels(xctl(), pkg, args[0])
 	}
-	ctl.xwrite(args[0])
+}
+
+func ctlcmdLoglevels(ctl *ctl) {
+	ctl.xwrite("loglevels")
+	ctl.xreadok()
+	ctl.xstreamto(os.Stdout)
+}
+
+func ctlcmdSetLoglevels(ctl *ctl, pkg, level string) {
+	ctl.xwrite("setloglevels")
+	ctl.xwrite(pkg)
+	ctl.xwrite(level)
 	ctl.xreadok()
 }
 
@@ -934,6 +997,70 @@ new mail deliveries.
 	fmt.Println("mox stopped")
 }
 
+func cmdBackup(c *cmd) {
+	c.params = "dest-dir"
+	c.help = `Creates a backup of the data directory.
+
+Backup creates consistent snapshots of the databases and message files and
+copies other files in the data directory. Empty directories are not copied.
+These files can then be stored elsewhere for long-term storage, or used to fall
+back to should an upgrade fail. Simply copying files in the data directory
+while mox is running can result in unusable database files.
+
+Message files never change (they are read-only, though can be removed) and are
+hardlinked so they don't consume additional space. If hardlinking fails, for
+example when the backup destination directory is on a different file system, a
+regular copy is made. Using a destination directory like "data/tmp/backup"
+increases the odds hardlinking succeeds: the default systemd service file
+specifically mounts the data directory, causing attempts to hardlink outside it
+to fail with an error about cross-device linking.
+
+All files in the data directory that aren't recognized (i.e. other than known
+database files, message files, an acme directory, the "tmp" directory, etc),
+are stored, but with a warning.
+
+A clean successful backup does not print any output by default. Use the
+-verbose flag for details, including timing.
+
+To restore a backup, first shut down mox, move away the old data directory and
+move an earlier backed up directory in its place, run "mox verifydata",
+possibly with the "-fix" option, and restart mox. After the restore, you may
+also want to run "mox bumpuidvalidity" for each account for which messages in a
+mailbox changed, to force IMAP clients to synchronize mailbox state.
+
+Before upgrading, to check if the upgrade will likely succeed, first make a
+backup, then use the new mox binary to run "mox verifydata" on the backup. This
+can change the backup files (e.g. upgrade database files, move away
+unrecognized message files), so you should make a new backup before actually
+upgrading.
+`
+
+	var verbose bool
+	c.flag.BoolVar(&verbose, "verbose", false, "print progress")
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+	mustLoadConfig()
+
+	dstDataDir, err := filepath.Abs(args[0])
+	xcheckf(err, "making path absolute")
+
+	ctlcmdBackup(xctl(), dstDataDir, verbose)
+}
+
+func ctlcmdBackup(ctl *ctl, dstDataDir string, verbose bool) {
+	ctl.xwrite("backup")
+	ctl.xwrite(dstDataDir)
+	if verbose {
+		ctl.xwrite("verbose")
+	} else {
+		ctl.xwrite("")
+	}
+	ctl.xstreamto(os.Stdout)
+	ctl.xreadok()
+}
+
 func cmdSetadminpassword(c *cmd) {
 	c.help = `Set a new admin password, for the web interface.
 
@@ -965,7 +1092,7 @@ WARNING: Bots will try to bruteforce your password. Connections with failed
 authentication attempts will be rate limited but attackers WILL find weak
 passwords. If your account is compromised, spammers are likely to abuse your
 system, spamming your address and the wider internet in your name. So please
-pick a random, unguessable password, preferrably at least 12 characters.
+pick a random, unguessable password, preferably at least 12 characters.
 
 `)
 	fmt.Printf("password: ")
@@ -999,10 +1126,13 @@ Any email address configured for the account can be used.
 
 	pw := xreadpassword()
 
-	ctl := xctl()
+	ctlcmdSetaccountpassword(xctl(), args[0], pw)
+}
+
+func ctlcmdSetaccountpassword(ctl *ctl, address, password string) {
 	ctl.xwrite("setaccountpassword")
-	ctl.xwrite(args[0])
-	ctl.xwrite(pw)
+	ctl.xwrite(address)
+	ctl.xwrite(password)
 	ctl.xreadok()
 }
 
@@ -1015,10 +1145,12 @@ func cmdDeliver(c *cmd) {
 		c.Usage()
 	}
 	mustLoadConfig()
+	ctlcmdDeliver(xctl(), args[0])
+}
 
-	ctl := xctl()
+func ctlcmdDeliver(ctl *ctl, address string) {
 	ctl.xwrite("deliver")
-	ctl.xwrite(args[0])
+	ctl.xwrite(address)
 	ctl.xreadok()
 	ctl.xstreamfrom(os.Stdin)
 	line := ctl.xread()
@@ -1039,8 +1171,10 @@ error.
 		c.Usage()
 	}
 	mustLoadConfig()
+	ctlcmdQueueList(xctl())
+}
 
-	ctl := xctl()
+func ctlcmdQueueList(ctl *ctl) {
 	ctl.xwrite("queue")
 	ctl.xreadok()
 	if _, err := io.Copy(os.Stdout, ctl.reader()); err != nil {
@@ -1049,29 +1183,37 @@ error.
 }
 
 func cmdQueueKick(c *cmd) {
-	c.params = "[-id id] [-todomain domain] [-recipient address]"
+	c.params = "[-id id] [-todomain domain] [-recipient address] [-transport transport]"
 	c.help = `Schedule matching messages in the queue for immediate delivery.
 
 Messages deliveries are normally attempted with exponential backoff. The first
 retry after 7.5 minutes, and doubling each time. Kicking messages sets their
 next scheduled attempt to now, it can cause delivery to fail earlier than
 without rescheduling.
+
+With the -transport flag, future delivery attempts are done using the specified
+transport. Transports can be configured in mox.conf, e.g. to submit to a remote
+queue over SMTP.
 `
 	var id int64
-	var todomain, recipient string
+	var todomain, recipient, transport string
 	c.flag.Int64Var(&id, "id", 0, "id of message in queue")
 	c.flag.StringVar(&todomain, "todomain", "", "destination domain of messages")
 	c.flag.StringVar(&recipient, "recipient", "", "recipient email address")
+	c.flag.StringVar(&transport, "transport", "", "transport to use for the next delivery")
 	if len(c.Parse()) != 0 {
 		c.Usage()
 	}
 	mustLoadConfig()
+	ctlcmdQueueKick(xctl(), id, todomain, recipient, transport)
+}
 
-	ctl := xctl()
+func ctlcmdQueueKick(ctl *ctl, id int64, todomain, recipient, transport string) {
 	ctl.xwrite("queuekick")
 	ctl.xwrite(fmt.Sprintf("%d", id))
 	ctl.xwrite(todomain)
 	ctl.xwrite(recipient)
+	ctl.xwrite(transport)
 	count := ctl.xread()
 	line := ctl.xread()
 	if line == "ok" {
@@ -1097,8 +1239,10 @@ the message, use "queue dump" before removing.
 		c.Usage()
 	}
 	mustLoadConfig()
+	ctlcmdQueueDrop(xctl(), id, todomain, recipient)
+}
 
-	ctl := xctl()
+func ctlcmdQueueDrop(ctl *ctl, id int64, todomain, recipient string) {
 	ctl.xwrite("queuedrop")
 	ctl.xwrite(fmt.Sprintf("%d", id))
 	ctl.xwrite(todomain)
@@ -1123,10 +1267,12 @@ The message is printed to stdout and is in standard internet mail format.
 		c.Usage()
 	}
 	mustLoadConfig()
+	ctlcmdQueueDump(xctl(), args[0])
+}
 
-	ctl := xctl()
+func ctlcmdQueueDump(ctl *ctl, id string) {
 	ctl.xwrite("queuedump")
-	ctl.xwrite(args[0])
+	ctl.xwrite(id)
 	ctl.xreadok()
 	if _, err := io.Copy(os.Stdout, ctl.reader()); err != nil {
 		log.Fatalf("%s", err)
@@ -1272,6 +1418,51 @@ that was passed.
 	}
 }
 
+func cmdDKIMSign(c *cmd) {
+	c.params = "message"
+	c.help = `Sign a message, adding DKIM-Signature headers based on the domain in the From header.
+
+The message is parsed, the domain looked up in the configuration files, and
+DKIM-Signature headers generated. The message is printed with the DKIM-Signature
+headers prepended.
+`
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+
+	msgf, err := os.Open(args[0])
+	xcheckf(err, "open message")
+	defer msgf.Close()
+
+	p, err := message.Parse(msgf)
+	xcheckf(err, "parsing message")
+
+	if len(p.Envelope.From) != 1 {
+		log.Fatalf("found %d from headers, need exactly 1", len(p.Envelope.From))
+	}
+	localpart := smtp.Localpart(p.Envelope.From[0].User)
+	dom, err := dns.ParseDomain(p.Envelope.From[0].Host)
+	xcheckf(err, "parsing domain in from header")
+
+	mustLoadConfig()
+
+	domConf, ok := mox.Conf.Domain(dom)
+	if !ok {
+		log.Fatalf("domain %s not configured", dom)
+	}
+
+	headers, err := dkim.Sign(context.Background(), localpart, dom, domConf.DKIM, false, msgf)
+	xcheckf(err, "signing message with dkim")
+	if headers == "" {
+		log.Fatalf("no DKIM configured for domain %s", dom)
+	}
+	_, err = fmt.Fprint(os.Stdout, headers)
+	xcheckf(err, "write headers")
+	_, err = io.Copy(os.Stdout, msgf)
+	xcheckf(err, "write message")
+}
+
 func cmdDKIMLookup(c *cmd) {
 	c.params = "selector domain"
 	c.help = "Lookup and print the DKIM record for the selector at the domain."
@@ -1382,7 +1573,7 @@ can be found in message headers.
 			} else {
 				spfIdentity = heloDomain
 			}
-			fmt.Printf("spf result: %q: %q\n", spfDomain, spfStatus)
+			fmt.Printf("spf result: %s: %s\n", spfDomain, spfStatus)
 		}
 	}
 
@@ -1632,9 +1823,12 @@ implementation has changed.
 	}
 
 	mustLoadConfig()
-	ctl := xctl()
+	ctlcmdRetrain(xctl(), args[0])
+}
+
+func ctlcmdRetrain(ctl *ctl, account string) {
 	ctl.xwrite("retrain")
-	ctl.xwrite(args[0])
+	ctl.xwrite(account)
 	ctl.xreadok()
 }
 
@@ -1807,7 +2001,7 @@ func cmdEnsureParsed(c *cmd) {
 	}()
 
 	n := 0
-	err = a.DB.Write(func(tx *bstore.Tx) error {
+	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
 		q := bstore.QueryTx[store.Message](tx)
 		q.FilterFn(func(m store.Message) bool {
 			return all || m.ParsedBuf == nil
@@ -1837,12 +2031,36 @@ func cmdEnsureParsed(c *cmd) {
 	fmt.Printf("%d messages updated\n", n)
 }
 
+func cmdMessageParse(c *cmd) {
+	c.unlisted = true
+	c.params = "message.eml"
+	c.help = "Parse message, print JSON representation."
+
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+
+	f, err := os.Open(args[0])
+	xcheckf(err, "open")
+	defer f.Close()
+
+	part, err := message.Parse(f)
+	xcheckf(err, "parsing message")
+	err = part.Walk(nil)
+	xcheckf(err, "parsing nested parts")
+	enc := json.NewEncoder(os.Stdout)
+	enc.SetIndent("", "\t")
+	err = enc.Encode(part)
+	xcheckf(err, "write")
+}
+
 func cmdBumpUIDValidity(c *cmd) {
 	c.unlisted = true
-	c.params = "account mailbox"
+	c.params = "account [mailbox]"
 	c.help = "Change the IMAP UID validity of the mailbox, causing IMAP clients to refetch messages."
 	args := c.Parse()
-	if len(args) != 2 {
+	if len(args) != 1 && len(args) != 2 {
 		c.Usage()
 	}
 
@@ -1856,261 +2074,190 @@ func cmdBumpUIDValidity(c *cmd) {
 	}()
 
 	var uidvalidity uint32
-	err = a.DB.Write(func(tx *bstore.Tx) error {
-		mb, err := bstore.QueryTx[store.Mailbox](tx).FilterEqual("Name", args[1]).Get()
+	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
+		q := bstore.QueryTx[store.Mailbox](tx)
+		if len(args) == 2 {
+			q.FilterEqual("Name", args[1])
+		}
+		mbl, err := q.SortAsc("Name").List()
 		if err != nil {
 			return fmt.Errorf("looking up mailbox: %v", err)
 		}
-		mb.UIDValidity++
-		uidvalidity = mb.UIDValidity
-		err = tx.Update(&mb)
-		if err != nil {
-			return fmt.Errorf("updating uid validity for mailbox: %v", err)
+		if len(args) == 2 && len(mbl) != 1 {
+			return fmt.Errorf("looking up mailbox %q, found %d mailboxes", args[1], len(mbl))
+		}
+		for _, mb := range mbl {
+			mb.UIDValidity++
+			uidvalidity = mb.UIDValidity
+			err = tx.Update(&mb)
+			if err != nil {
+				return fmt.Errorf("updating uid validity for mailbox: %v", err)
+			}
+			fmt.Printf("uid validity for %q updated to %d\n", mb.Name, uidvalidity)
 		}
 		return nil
 	})
-	xcheckf(err, "updating uidvalidity for mailbox: %v", err)
-	fmt.Printf("uid validity for %q is now %d\n", args[1], uidvalidity)
+	xcheckf(err, "updating database")
 }
 
-var submitconf struct {
-	LocalHostname      string `sconf-doc:"Hosts don't always have an FQDN, set it explicitly, for EHLO."`
-	Host               string `sconf-doc:"Host to dial for delivery, e.g. mail.<domain>."`
-	Port               int    `sconf-doc:"Port to dial for delivery, e.g. 465 for submissions, 587 for submission, or perhaps 25 for smtp."`
-	TLS                bool   `sconf-doc:"Connect with TLS. Usually for connections to port 465."`
-	STARTTLS           bool   `sconf-doc:"After starting in plain text, use STARTTLS to enable TLS. For port 587 and 25."`
-	Username           string `sconf-doc:"For SMTP plain auth."`
-	Password           string `sconf-doc:"For SMTP plain auth."`
-	AuthMethod         string `sconf-doc:"Ignored for now, regardless of value, AUTH PLAIN is done. This will change in the future."`
-	From               string `sconf-doc:"Address for MAIL FROM in SMTP and From-header in message."`
-	DefaultDestination string `sconf:"optional" sconf-doc:"Used when specified address does not contain an @ and may be a local user (eg root)."`
-}
+func cmdReassignUIDs(c *cmd) {
+	c.unlisted = true
+	c.params = "account [mailboxid]"
+	c.help = `Reassign UIDs in one mailbox or all mailboxes in an account and bump UID validity, causing IMAP clients to refetch messages.
 
-func cmdConfigDescribeSendmail(c *cmd) {
-	c.params = ">/etc/moxsubmit.conf"
-	c.help = `Describe configuration for mox when invoked as sendmail.`
-	if len(c.Parse()) != 0 {
+Opens account database file directly. Ensure mox does not have the account
+open, or is not running.
+`
+	args := c.Parse()
+	if len(args) != 1 && len(args) != 2 {
 		c.Usage()
 	}
 
-	err := sconf.Describe(os.Stdout, submitconf)
-	xcheckf(err, "describe config")
+	var mailboxID int64
+	if len(args) == 2 {
+		var err error
+		mailboxID, err = strconv.ParseInt(args[1], 10, 64)
+		xcheckf(err, "parsing mailbox id")
+	}
+
+	mustLoadConfig()
+	a, err := store.OpenAccount(args[0])
+	xcheckf(err, "open account")
+	defer func() {
+		if err := a.Close(); err != nil {
+			log.Printf("closing account: %v", err)
+		}
+	}()
+
+	// Gather the last-assigned UIDs per mailbox.
+	uidlasts := map[int64]store.UID{}
+
+	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
+		// Reassign UIDs, going per mailbox. We assign starting at 1, only changing the
+		// message if it isn't already at the intended UID. Doing it in this order ensures
+		// we don't get into trouble with duplicate UIDs for a mailbox.
+		q := bstore.QueryTx[store.Message](tx)
+		if len(args) == 2 {
+			q.FilterNonzero(store.Message{MailboxID: mailboxID})
+		}
+		q.SortAsc("MailboxID", "UID")
+		err := q.ForEach(func(m store.Message) error {
+			uidlasts[m.MailboxID]++
+			uid := uidlasts[m.MailboxID]
+			if m.UID != uid {
+				m.UID = uid
+				if err := tx.Update(&m); err != nil {
+					return fmt.Errorf("updating uid for message: %v", err)
+				}
+			}
+			return nil
+		})
+		if err != nil {
+			return fmt.Errorf("reading through messages: %v", err)
+		}
+
+		// Now update the uidnext and uidvalidity for each mailbox.
+		err = bstore.QueryTx[store.Mailbox](tx).ForEach(func(mb store.Mailbox) error {
+			// Assign each mailbox a completely new uidvalidity.
+			uidvalidity, err := a.NextUIDValidity(tx)
+			if err != nil {
+				return fmt.Errorf("assigning next uid validity: %v", err)
+			}
+
+			if mb.UIDValidity >= uidvalidity {
+				// This should not happen, but since we're fixing things up after a hypothetical
+				// mishap, might as well account for inconsistent uidvalidity.
+				next := store.NextUIDValidity{ID: 1, Next: mb.UIDValidity + 2}
+				if err := tx.Update(&next); err != nil {
+					log.Printf("updating nextuidvalidity: %v, continuing", err)
+				}
+				mb.UIDValidity++
+			} else {
+				mb.UIDValidity = uidvalidity
+			}
+			mb.UIDNext = uidlasts[mb.ID] + 1
+			if err := tx.Update(&mb); err != nil {
+				return fmt.Errorf("updating uidvalidity and uidnext for mailbox: %v", err)
+			}
+			return nil
+		})
+		if err != nil {
+			return fmt.Errorf("updating mailboxes: %v", err)
+		}
+		return nil
+	})
+	xcheckf(err, "updating database")
 }
 
-func cmdSendmail(c *cmd) {
-	c.params = "[-Fname] [ignoredflags] [-t] [<message]"
-	c.help = `Sendmail is a drop-in replacement for /usr/sbin/sendmail to deliver emails sent by unix processes like cron.
+func cmdFixUIDMeta(c *cmd) {
+	c.unlisted = true
+	c.params = "account"
+	c.help = `Fix inconsistent UIDVALIDITY and UIDNEXT in messages/mailboxes/account.
 
-If invoked as "sendmail", it will act as sendmail for sending messages. Its
-intention is to let processes like cron send emails. Messages are submitted to
-an actual mail server over SMTP. The destination mail server and credentials are
-configured in /etc/moxsubmit.conf, see mox config describe-sendmail. The From
-message header is rewritten to the configured address. When the addressee
-appears to be a local user, because without @, the message is sent to the
-configured default address.
+The next UID to use for a message in a mailbox should always be higher than any
+existing message UID in the mailbox. If it is not, the mailbox UIDNEXT is
+updated.
 
-If submitting an email fails, it is added to a directory moxsubmit.failures in
-the user's home directory.
+Each mailbox has a UIDVALIDITY sequence number, which should always be lower
+than the per-account next UIDVALIDITY to use. If it is not, the account next
+UIDVALIDITY is updated.
 
-Most flags are ignored to fake compatibility with other sendmail
-implementations. A single recipient is required, or the tflag.
-
-/etc/moxsubmit.conf should be group-readable and not readable by others and this
-binary should be setgid that group:
-
-	groupadd moxsubmit
-	install -m 2755 -o root -g moxsubmit mox /usr/sbin/sendmail
-	touch /etc/moxsubmit.conf
-	chown root:moxsubmit /etc/moxsubmit.conf
-	chmod 640 /etc/moxsubmit.conf
-	# edit /etc/moxsubmit.conf
+Opens account database file directly. Ensure mox does not have the account
+open, or is not running.
 `
-
-	// We are faking that we parse flags, this is non-standard, we want to be lax and ignore most flags.
-	args := c.flagArgs
-	c.flagArgs = []string{}
-	c.Parse() // We still have to call Parse for the usage gathering.
-
-	// Typical cron usage of sendmail:
-	// anacron: https://salsa.debian.org/debian/anacron/-/blob/c939c8c80fc9419c11a5e6be5cbe84f03ad332fd/runjob.c#L183
-	// cron: https://github.com/vixie/cron/blob/fea7a6c5421f88f034be8eef66a84d8b65b5fbe0/config.h#L41
-
-	var from string
-	var tflag bool // If set, we need to take the recipient(s) from the message headers. We only do one recipient, in To.
-	o := 0
-	for i, s := range args {
-		if s == "--" {
-			o = i + 1
-			break
-		}
-		if !strings.HasPrefix(s, "-") {
-			o = i
-			break
-		}
-		s = s[1:]
-		if strings.HasPrefix(s, "F") {
-			from = s[1:]
-			log.Printf("ignoring -F %q", from) // todo
-		} else if s == "t" {
-			tflag = true
-		}
-		o = i + 1
-		// Ignore options otherwise.
-		// todo: we may want to parse more flags. some invocations may not be about sending a message. for now, we'll assume sendmail is only invoked to send a message.
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
 	}
-	args = args[o:]
 
-	// todo: perhaps allow configuration of config file through environment variable? have to keep in mind that mox with setgid moxsubmit would be reading the file.
-	const confPath = "/etc/moxsubmit.conf"
-	err := sconf.ParseFile(confPath, &submitconf)
-	xcheckf(err, "parsing config")
+	mustLoadConfig()
+	a, err := store.OpenAccount(args[0])
+	xcheckf(err, "open account")
+	defer func() {
+		if err := a.Close(); err != nil {
+			log.Printf("closing account: %v", err)
+		}
+	}()
 
-	var recipient string
-	if len(args) == 1 && !tflag {
-		recipient = args[0]
-		if !strings.Contains(recipient, "@") {
-			if submitconf.DefaultDestination == "" {
-				log.Fatalf("recipient %q has no @ and no default destination configured", recipient)
+	var maxUIDValidity uint32
+
+	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
+		// We look at each mailbox, retrieve its max UID and compare against the mailbox
+		// UIDNEXT.
+		err := bstore.QueryTx[store.Mailbox](tx).ForEach(func(mb store.Mailbox) error {
+			if mb.UIDValidity > maxUIDValidity {
+				maxUIDValidity = mb.UIDValidity
 			}
-			recipient = submitconf.DefaultDestination
-		} else {
-			_, err := smtp.ParseAddress(args[0])
-			xcheckf(err, "parsing recipient address")
-		}
-	} else if !tflag || len(args) != 0 {
-		log.Fatalln("need either exactly 1 recipient, or -t")
-	}
-
-	// Read message and build message we are going to send. We replace \n
-	// with \r\n, and we replace the From header.
-	// todo: should we also wrap lines that are too long? perhaps only if this is just text, no multipart?
-	var sb strings.Builder
-	r := bufio.NewReader(os.Stdin)
-	header := true // Whether we are in the header.
-	fmt.Fprintf(&sb, "From: <%s>\r\n", submitconf.From)
-	var haveTo bool
-	for {
-		line, err := r.ReadString('\n')
-		if err != nil && err != io.EOF {
-			xcheckf(err, "reading message")
-		}
-		if line != "" {
-			if !strings.HasSuffix(line, "\n") {
-				line += "\n"
+			m, err := bstore.QueryTx[store.Message](tx).FilterNonzero(store.Message{MailboxID: mb.ID}).SortDesc("UID").Limit(1).Get()
+			if err == bstore.ErrAbsent || err == nil && m.UID < mb.UIDNext {
+				return nil
+			} else if err != nil {
+				return fmt.Errorf("finding message with max uid in mailbox: %w", err)
 			}
-			if !strings.HasSuffix(line, "\r\n") {
-				line = line[:len(line)-1] + "\r\n"
+			mb.UIDNext = m.UID + 1
+			if err := tx.Update(&mb); err != nil {
+				log.Printf("fixing uidnext to %d (max uid is %d) for mailbox id %d", mb.UIDNext, m.UID, mb.ID)
+				return fmt.Errorf("updating mailbox uidnext: %v", err)
 			}
-			if header && line == "\r\n" {
-				// Bare \r\n marks end of header.
-				if !haveTo {
-					line = fmt.Sprintf("To: <%s>\r\n", recipient) + line
-				}
-				header = false
-			} else if header {
-				t := strings.SplitN(line, ":", 2)
-				if len(t) != 2 {
-					log.Fatalf("invalid message, missing colon in header")
-				}
-				k := strings.ToLower(t[0])
-				if k == "from" {
-					// We already added a From header.
-					if err == io.EOF {
-						break
-					}
-					continue
-				} else if tflag && k == "to" {
-					if recipient != "" {
-						log.Fatalf("only single To header allowed")
-					}
-					s := strings.TrimSpace(t[1])
-					if !strings.Contains(s, "@") {
-						if submitconf.DefaultDestination == "" {
-							log.Fatalf("recipient %q has no @ and no default destination is configured", s)
-						}
-						recipient = submitconf.DefaultDestination
-					} else {
-						addrs, err := mail.ParseAddressList(s)
-						xcheckf(err, "parsing To address list")
-						if len(addrs) != 1 {
-							log.Fatalf("only single address allowed in To header")
-						}
-						recipient = addrs[0].Address
-					}
-				}
-				if k == "to" {
-					haveTo = true
-				}
+			return nil
+		})
+		if err != nil {
+			return fmt.Errorf("processing mailboxes: %v", err)
+		}
+
+		uidvalidity := store.NextUIDValidity{ID: 1}
+		if err := tx.Get(&uidvalidity); err != nil {
+			return fmt.Errorf("reading account next uidvalidity: %v", err)
+		}
+		if maxUIDValidity >= uidvalidity.Next {
+			log.Printf("account next uidvalidity %d <= highest uidvalidity %d found in mailbox, resetting account next uidvalidity to %d", uidvalidity.Next, maxUIDValidity, maxUIDValidity+1)
+			uidvalidity.Next = maxUIDValidity + 1
+			if err := tx.Update(&uidvalidity); err != nil {
+				return fmt.Errorf("updating account next uidvalidity: %v", err)
 			}
-			sb.WriteString(line)
 		}
-		if err == io.EOF {
-			break
-		}
-	}
-	msg := sb.String()
 
-	if recipient == "" {
-		log.Fatalf("no recipient")
-	}
-
-	// Message seems acceptable. We'll try to deliver it from here. If that fails, we
-	// store the message in the users home directory.
-
-	xcheckf := func(err error, format string, args ...any) {
-		if err == nil {
-			return
-		}
-		log.Printf("submit failed: %s: %s", fmt.Sprintf(format, args...), err)
-		homedir, err := os.UserHomeDir()
-		xcheckf(err, "finding homedir for storing message after failed delivery")
-		maildir := filepath.Join(homedir, "moxsubmit.failures")
-		os.Mkdir(maildir, 0700)
-		f, err := os.CreateTemp(maildir, "newmsg.")
-		xcheckf(err, "creating temp file for storing message after failed delivery")
-		defer func() {
-			if f != nil {
-				if err := os.Remove(f.Name()); err != nil {
-					log.Printf("removing temp file after failure storing failed delivery: %v", err)
-				}
-			}
-		}()
-		_, err = f.Write([]byte(msg))
-		xcheckf(err, "writing message to temp file after failed delivery")
-		name := f.Name()
-		err = f.Close()
-		xcheckf(err, "closing message in temp file after failed delivery")
-		f = nil
-		log.Printf("saved message in %s", name)
-		os.Exit(1)
-	}
-
-	var conn net.Conn
-	addr := net.JoinHostPort(submitconf.Host, fmt.Sprintf("%d", submitconf.Port))
-	d := net.Dialer{Timeout: 30 * time.Second}
-	if submitconf.TLS {
-		conn, err = tls.DialWithDialer(&d, "tcp", addr, nil)
-	} else {
-		conn, err = d.Dial("tcp", addr)
-	}
-	xcheckf(err, "dial submit server")
-
-	ctx, cancel := context.WithTimeout(context.Background(), 3*time.Minute)
-	defer cancel()
-	tlsMode := smtpclient.TLSStrict
-	if !submitconf.STARTTLS {
-		tlsMode = smtpclient.TLSSkip
-	}
-	// todo: should have more auth options, scram-sha-256 at least, perhaps cram-md5 for compatibility as well.
-	authLine := fmt.Sprintf("AUTH PLAIN %s", base64.StdEncoding.EncodeToString([]byte(fmt.Sprintf("\u0000%s\u0000%s", submitconf.Username, submitconf.Password))))
-	mox.Conf.Static.HostnameDomain.ASCII = submitconf.LocalHostname
-	client, err := smtpclient.New(ctx, mlog.New("sendmail"), conn, tlsMode, submitconf.Host, authLine)
-	xcheckf(err, "open smtp session")
-
-	err = client.Deliver(ctx, submitconf.From, recipient, int64(len(msg)), strings.NewReader(msg), true, false)
-	xcheckf(err, "submit message")
-
-	if err := client.Close(); err != nil {
-		log.Printf("closing smtp session after message was sent: %v", err)
-	}
+		return nil
+	})
+	xcheckf(err, "updating database")
 }
