@@ -125,16 +125,16 @@ webhandlers".
 	if err != nil {
 		fatalf("parsing email address: %s", err)
 	}
-	username := addr.Localpart.String()
+	accountName := addr.Localpart.String()
 	domain := addr.Domain
 
-	for _, c := range username {
+	for _, c := range accountName {
 		if c > 0x7f {
 			fmt.Printf(`NOTE: Username %q is not ASCII-only. It is recommended you also configure an
 ASCII-only alias. Both for delivery of email from other systems, and for
 logging in with IMAP.
 
-`, username)
+`, accountName)
 			break
 		}
 	}
@@ -305,7 +305,6 @@ again with the -hostname flag.
 		}
 	}
 
-	// todo: lookup without going through /etc/hosts, because a machine typically has its name configured there, and LookupIPAddr will return it, but we care about DNS settings that the rest of the world uses to find us. perhaps we should check if the address resolves to 127.0.0.0/8?
 	fmt.Printf("Looking up IPs for hostname %s...", dnshostname)
 	ipctx, ipcancel := context.WithTimeout(resolveCtx, 5*time.Second)
 	defer ipcancel()
@@ -313,17 +312,22 @@ again with the -hostname flag.
 	ipcancel()
 	var xips []net.IPAddr
 	var xipstrs []string
+	var dnswarned bool
 	for _, ip := range ips {
 		// During linux install, you may get an alias for you full hostname in /etc/hosts
 		// resolving to 127.0.1.1, which would result in a false positive about the
 		// hostname having a record. Filter it out. It is a bit surprising that hosts don't
 		// otherwise know their FQDN.
-		if !ip.IP.IsLoopback() {
-			xips = append(xips, ip)
-			xipstrs = append(xipstrs, ip.String())
+		if ip.IP.IsLoopback() {
+			dnswarned = true
+			fmt.Printf("\n\nWARNING: Your hostname is resolving to a loopback IP address %s. This likely breaks email delivery to local accounts. /etc/hosts likely contains a line like %q. Either replace it with your actual IP(s), or remove the line.\n", ip.IP, fmt.Sprintf("%s %s", ip.IP, dnshostname.ASCII))
+			continue
 		}
+		xips = append(xips, ip)
+		xipstrs = append(xipstrs, ip.String())
 	}
 	if err == nil && len(xips) == 0 {
+		// todo: possibly check this by trying to resolve without using /etc/hosts?
 		err = errors.New("hostname not in dns, probably only in /etc/hosts")
 	}
 	ips = xips
@@ -333,8 +337,11 @@ again with the -hostname flag.
 		publicIPs = xipstrs
 	}
 	if err != nil {
+		if !dnswarned {
+			fmt.Printf("\n")
+		}
+		dnswarned = true
 		fmt.Printf(`
-
 WARNING: Quickstart assumed the hostname of this machine is %s and generates a
 config for that host, but could not retrieve that name from DNS:
 
@@ -351,7 +358,9 @@ This likely means one of two things:
 
 
 `, dnshostname, err)
-	} else {
+	}
+
+	if !dnswarned {
 		fmt.Printf(" OK\n")
 
 		var l []string
@@ -538,7 +547,7 @@ listed in more DNS block lists, visit:
 		"public":   public,
 		"internal": internal,
 	}
-	sc.Postmaster.Account = username
+	sc.Postmaster.Account = accountName
 	sc.Postmaster.Mailbox = "Postmaster"
 
 	mox.ConfigStaticPath = "config/mox.conf"
@@ -548,7 +557,7 @@ listed in more DNS block lists, visit:
 
 	accountConf := mox.MakeAccountConfig(addr)
 	const withMTASTS = true
-	confDomain, keyPaths, err := mox.MakeDomainConfig(context.Background(), domain, dnshostname, username, withMTASTS)
+	confDomain, keyPaths, err := mox.MakeDomainConfig(context.Background(), domain, dnshostname, accountName, withMTASTS)
 	if err != nil {
 		fatalf("making domain config: %s", err)
 	}
@@ -558,7 +567,7 @@ listed in more DNS block lists, visit:
 		domain.Name(): confDomain,
 	}
 	dc.Accounts = map[string]config.Account{
-		username: accountConf,
+		accountName: accountConf,
 	}
 
 	// Build config in memory, so we can easily comment out the DNSBLs config.
@@ -584,12 +593,12 @@ listed in more DNS block lists, visit:
 	// This approach is a bit horrible, but it generates a convenient
 	// example that includes the comments. Though it is gone by the first
 	// write of the file by mox.
-	odests := fmt.Sprintf("\t\tDestinations:\n\t\t\t%s: nil\n", addr.Localpart.String())
+	odests := fmt.Sprintf("\t\tDestinations:\n\t\t\t%s: nil\n", addr.String())
 	var destsExample = struct {
 		Destinations map[string]config.Destination
 	}{
 		Destinations: map[string]config.Destination{
-			addr.Localpart.String(): {
+			addr.String(): {
 				Rulesets: []config.Ruleset{
 					{
 						VerifiedDomain: "list.example.org",
@@ -641,7 +650,7 @@ listed in more DNS block lists, visit:
 	if err != nil {
 		fatalf("open account: %s", err)
 	}
-	cleanupPaths = append(cleanupPaths, dataDir, filepath.Join(dataDir, "accounts"), filepath.Join(dataDir, "accounts", username), filepath.Join(dataDir, "accounts", username, "index.db"))
+	cleanupPaths = append(cleanupPaths, dataDir, filepath.Join(dataDir, "accounts"), filepath.Join(dataDir, "accounts", accountName), filepath.Join(dataDir, "accounts", accountName, "index.db"))
 
 	password := pwgen()
 	if err := acc.SetPassword(password); err != nil {
@@ -650,10 +659,10 @@ listed in more DNS block lists, visit:
 	if err := acc.Close(); err != nil {
 		fatalf("closing account: %s", err)
 	}
-	fmt.Printf("IMAP and SMTP submission password for %s: %s\n\n", args[0], password)
-	fmt.Println(`When configuring your email client, use the email address as username. If
-autoconfig/autodiscover does not work, use the settings below.`)
-	fmt.Println("")
+	fmt.Printf("IMAP, SMTP submission and HTTP account password for %s: %s\n\n", args[0], password)
+	fmt.Printf(`When configuring your email client, use the email address as username. If
+autoconfig/autodiscover does not work, use these settings:
+`)
 	printClientConfig(domain)
 
 	if existingWebserver {
@@ -739,11 +748,20 @@ starting up. On linux, you may want to enable mox as a systemd service.
 	sudo systemctl enable $PWD/mox.service
 	sudo systemctl start mox.service
 	sudo journalctl -f -u mox.service # See logs
-
 `)
 	}
 
-	fmt.Printf(`For secure email exchange you should have a strictly validating DNSSEC
+	fmt.Printf(`
+After starting mox, the web interfaces are served at:
+
+http://localhost/       - account (email address as username)
+http://localhost/admin/ - admin (empty username)
+
+To access these from your browser, run
+"ssh -L 8080:localhost:80 you@yourmachine" locally and open
+http://localhost:8080/[...].
+
+For secure email exchange you should have a strictly validating DNSSEC
 resolver. An easy and the recommended way is to install unbound.
 
 If you run into problem, have questions/feedback or found a bug, please let us
