@@ -80,7 +80,7 @@ func monitorDNSBL(log *mlog.Log) {
 		time.Sleep(sleep)
 		sleep = 3 * time.Hour
 
-		ips, err := mox.IPs(mox.Context)
+		ips, err := mox.IPs(mox.Context, false)
 		if err != nil {
 			log.Errorx("listing ips for dnsbl monitor", err)
 			continue
@@ -107,7 +107,7 @@ func monitorDNSBL(log *mlog.Log) {
 							Name: "mox_dnsbl_ips_success",
 							Help: "DNSBL lookups to configured DNSBLs of our IPs.",
 							ConstLabels: prometheus.Labels{
-								"zone": zone.String(),
+								"zone": zone.LogString(),
 								"ip":   k.ip,
 							},
 						},
@@ -127,6 +127,7 @@ func monitorDNSBL(log *mlog.Log) {
 	}
 }
 
+// also see localserve.go, code is similar or even shared.
 func cmdServe(c *cmd) {
 	c.help = `Start mox, serving SMTP/IMAP/HTTPS.
 
@@ -146,19 +147,28 @@ requested, other TLS certificates are requested on demand.
 	mlog.SetConfig(mox.Conf.Log)
 
 	checkACMEHosts := os.Getuid() != 0
-	mox.MustLoadConfig(checkACMEHosts)
 
 	log := mlog.New("serve")
 
 	if os.Getuid() == 0 {
+		mox.MustLoadConfig(true, checkACMEHosts)
+
 		// No need to potentially start and keep multiple processes. As root, we just need
 		// to start the child process.
 		runtime.GOMAXPROCS(1)
 
-		log.Print("starting as root, initializing network listeners", mlog.Field("version", moxvar.Version), mlog.Field("pid", os.Getpid()))
+		moxconf, err := filepath.Abs(mox.ConfigStaticPath)
+		log.Check(err, "finding absolute mox.conf path")
+		domainsconf, err := filepath.Abs(mox.ConfigDynamicPath)
+		log.Check(err, "finding absolute domains.conf path")
+
+		log.Print("starting as root, initializing network listeners", mlog.Field("version", moxvar.Version), mlog.Field("pid", os.Getpid()), mlog.Field("moxconf", moxconf), mlog.Field("domainsconf", domainsconf))
 		if os.Getenv("MOX_SOCKETS") != "" {
 			log.Fatal("refusing to start as root with $MOX_SOCKETS set")
 		}
+		if os.Getenv("MOX_FILES") != "" {
+			log.Fatal("refusing to start as root with $MOX_FILES set")
+		}
 
 		if !mox.Conf.Static.NoFixPermissions {
 			// Fix permissions now that we have privilege to do so. Useful for update of v0.0.1
@@ -177,7 +187,8 @@ requested, other TLS certificates are requested on demand.
 		}
 	} else {
 		log.Print("starting as unprivileged user", mlog.Field("user", mox.Conf.Static.User), mlog.Field("uid", mox.Conf.Static.UID), mlog.Field("gid", mox.Conf.Static.GID), mlog.Field("pid", os.Getpid()))
-		mox.RestorePassedSockets()
+		mox.RestorePassedFiles()
+		mox.MustLoadConfig(true, checkACMEHosts)
 	}
 
 	syscall.Umask(syscall.Umask(007) | 007)
@@ -264,7 +275,7 @@ requested, other TLS certificates are requested on demand.
 
 			var cl string
 			for _, c := range changelog.Changes {
-				cl += "----\n\n" + strings.TrimSpace(c.Text)
+				cl += "----\n\n" + strings.TrimSpace(c.Text) + "\n\n"
 			}
 			cl += "----"
 
@@ -291,7 +302,7 @@ requested, other TLS certificates are requested on demand.
 				}
 			}()
 			m := &store.Message{Received: time.Now(), Flags: store.Flags{Flagged: true}}
-			n, err := fmt.Fprintf(f, "Date: %s\r\nSubject: mox %s available\r\n\r\nHi!\r\n\r\nVersion %s of mox is available, this is install is at %s.\r\n\r\nChanges:\r\n\r\n%s\r\n\r\nPlease report any issues at https://github.com/mjl-/mox, thanks!\r\n\r\nCheers,\r\nmox\r\n", time.Now().Format(message.RFC5322Z), latest, latest, current, strings.ReplaceAll(cl, "\n", "\r\n"))
+			n, err := fmt.Fprintf(f, "Date: %s\r\nSubject: mox %s available\r\n\r\nHi!\r\n\r\nVersion %s of mox is available, this install is at %s.\r\n\r\nChanges:\r\n\r\n%s\r\n\r\nRemember to make a backup with \"mox backup\" before upgrading.\r\nPlease report any issues at https://github.com/mjl-/mox, thanks!\r\n\r\nCheers,\r\nmox\r\n", time.Now().Format(message.RFC5322Z), latest, latest, current, strings.ReplaceAll(cl, "\n", "\r\n"))
 			if err != nil {
 				log.Infox("writing temporary message file for changelog delivery", err)
 				return next
@@ -320,38 +331,6 @@ requested, other TLS certificates are requested on demand.
 
 	go monitorDNSBL(log)
 
-	shutdown := func() {
-		// We indicate we are shutting down. Causes new connections and new SMTP commands
-		// to be rejected. Should stop active connections pretty quickly.
-		mox.ShutdownCancel()
-
-		// Now we are going to wait for all connections to be gone, up to a timeout.
-		done := mox.Connections.Done()
-		second := time.Tick(time.Second)
-		select {
-		case <-done:
-			log.Print("connections shutdown, waiting until 1 second passed")
-			<-second
-
-		case <-time.Tick(3 * time.Second):
-			// We now cancel all pending operations, and set an immediate deadline on sockets.
-			// Should get us a clean shutdown relatively quickly.
-			mox.ContextCancel()
-			mox.Connections.Shutdown()
-
-			second := time.Tick(time.Second)
-			select {
-			case <-done:
-				log.Print("no more connections, shutdown is clean, waiting until 1 second passed")
-				<-second // Still wait for second, giving processes like imports a chance to clean up.
-			case <-second:
-				log.Print("shutting down with pending sockets")
-			}
-		}
-		err := os.Remove(mox.DataDirPath("ctl"))
-		log.Check(err, "removing ctl unix domain socket during shutdown")
-	}
-
 	ctlpath := mox.DataDirPath("ctl")
 	_ = os.Remove(ctlpath)
 	ctl, err := net.Listen("unix", ctlpath)
@@ -367,7 +346,7 @@ requested, other TLS certificates are requested on demand.
 			}
 			cid := mox.Cid()
 			ctx := context.WithValue(mox.Context, mlog.CidKey, cid)
-			go servectl(ctx, log.WithCid(cid), conn, shutdown)
+			go servectl(ctx, log.WithCid(cid), conn, func() { shutdown(log) })
 		}
 	}()
 
@@ -398,7 +377,7 @@ requested, other TLS certificates are requested on demand.
 	signal.Notify(sigc, os.Interrupt, syscall.SIGTERM)
 	sig := <-sigc
 	log.Print("shutting down, waiting max 3s for existing connections", mlog.Field("signal", sig))
-	shutdown()
+	shutdown(log)
 	if num, ok := sig.(syscall.Signal); ok {
 		os.Exit(int(num))
 	} else {
@@ -406,6 +385,38 @@ requested, other TLS certificates are requested on demand.
 	}
 }
 
+func shutdown(log *mlog.Log) {
+	// We indicate we are shutting down. Causes new connections and new SMTP commands
+	// to be rejected. Should stop active connections pretty quickly.
+	mox.ShutdownCancel()
+
+	// Now we are going to wait for all connections to be gone, up to a timeout.
+	done := mox.Connections.Done()
+	second := time.Tick(time.Second)
+	select {
+	case <-done:
+		log.Print("connections shutdown, waiting until 1 second passed")
+		<-second
+
+	case <-time.Tick(3 * time.Second):
+		// We now cancel all pending operations, and set an immediate deadline on sockets.
+		// Should get us a clean shutdown relatively quickly.
+		mox.ContextCancel()
+		mox.Connections.Shutdown()
+
+		second := time.Tick(time.Second)
+		select {
+		case <-done:
+			log.Print("no more connections, shutdown is clean, waiting until 1 second passed")
+			<-second // Still wait for second, giving processes like imports a chance to clean up.
+		case <-second:
+			log.Print("shutting down with pending sockets")
+		}
+	}
+	err := os.Remove(mox.DataDirPath("ctl"))
+	log.Check(err, "removing ctl unix domain socket during shutdown")
+}
+
 // Set correct permissions for mox working directory, binary, config and data and service file.
 //
 // We require being able to stat the basic non-optional paths. Then we'll try to
@@ -583,7 +594,7 @@ func start(mtastsdbRefresher, skipForkExec bool) error {
 			mox.ForkExecUnprivileged()
 			panic("cannot happen")
 		} else {
-			mox.CleanupPassedSockets()
+			mox.CleanupPassedFiles()
 		}
 	}
 
