@@ -2,6 +2,7 @@ package smtpserver
 
 import (
 	"context"
+	"fmt"
 	"net"
 	"os"
 	"time"
@@ -160,7 +161,7 @@ func analyze(ctx context.Context, log *mlog.Log, resolver dns.Resolver, d delive
 	var reason string
 	var err error
 	d.acc.WithRLock(func() {
-		err = d.acc.DB.Read(func(tx *bstore.Tx) error {
+		err = d.acc.DB.Read(ctx, func(tx *bstore.Tx) error {
 			// Set message MailboxID to which mail will be delivered. Reputation is
 			// per-mailbox. If referenced mailbox is not found (e.g. does not yet exist), we
 			// can still determine a reputation because we also base it on outgoing
@@ -172,7 +173,10 @@ func analyze(ctx context.Context, log *mlog.Log, resolver dns.Resolver, d delive
 			if rs != nil {
 				mailbox = rs.Mailbox
 			}
-			mb := d.acc.MailboxFindX(tx, mailbox)
+			mb, err := d.acc.MailboxFind(tx, mailbox)
+			if err != nil {
+				return fmt.Errorf("finding destination mailbox: %w", err)
+			}
 			if mb != nil {
 				// We want to deliver to mb.ID, but this message may be rejected and sent to the
 				// Rejects mailbox instead, which MailboxID overwritten. Record the ID in
@@ -187,7 +191,6 @@ func analyze(ctx context.Context, log *mlog.Log, resolver dns.Resolver, d delive
 				log.Debug("mailbox not found in database", mlog.Field("mailbox", mailbox))
 			}
 
-			var err error
 			isjunk, conclusive, method, err = reputation(tx, log, d.m)
 			reason = string(method)
 			return err
@@ -248,13 +251,13 @@ func analyze(ctx context.Context, log *mlog.Log, resolver dns.Resolver, d delive
 	reason = reasonNoBadSignals
 	accept := true
 	var junkSubjectpass bool
-	f, jf, err := d.acc.OpenJunkFilter(log)
+	f, jf, err := d.acc.OpenJunkFilter(ctx, log)
 	if err == nil {
 		defer func() {
 			err := f.Close()
 			log.Check(err, "closing junkfilter")
 		}()
-		contentProb, _, _, _, err := f.ClassifyMessageReader(store.FileMsgReader(d.m.MsgPrefix, d.dataFile), d.m.Size)
+		contentProb, _, _, _, err := f.ClassifyMessageReader(ctx, store.FileMsgReader(d.m.MsgPrefix, d.dataFile), d.m.Size)
 		if err != nil {
 			log.Errorx("testing for spam", err)
 			return reject(smtp.C451LocalErr, smtp.SeSys3Other0, "error processing", err, reasonJunkClassifyError)
