@@ -20,7 +20,6 @@ database.
 package store
 
 // todo: make up a function naming scheme that indicates whether caller should broadcast changes.
-// todo: fewer (no?) "X" functions, but only explicit error handling.
 
 import (
 	"context"
@@ -40,6 +39,7 @@ import (
 	"time"
 
 	"golang.org/x/crypto/bcrypt"
+	"golang.org/x/exp/slices"
 	"golang.org/x/text/unicode/norm"
 
 	"github.com/mjl-/bstore"
@@ -173,6 +173,11 @@ type Mailbox struct {
 	Junk    bool
 	Sent    bool
 	Trash   bool
+
+	// Keywords as used in messages. Storing a non-system keyword for a message
+	// automatically adds it to this list. Used in the IMAP FLAGS response. Only
+	// "atoms", stored in lower case.
+	Keywords []string
 }
 
 // Subscriptions are separate from existence of mailboxes.
@@ -278,8 +283,7 @@ type Message struct {
 	MailFromValidation Validation // Can have SPF-specific validations like ValidationSoftfail.
 	MsgFromValidation  Validation // Desirable validations: Strict, DMARC, Relaxed. Will not be just Pass.
 
-	// todo: needs an "in" index, which bstore does not yet support. for performance while checking reputation.
-	DKIMDomains []string // Domains with verified DKIM signatures. Unicode string.
+	DKIMDomains []string `bstore:"index DKIMDomains+Received"` // Domains with verified DKIM signatures. Unicode string.
 
 	// Value of Message-Id header. Only set for messages that were
 	// delivered to the rejects mailbox. For ensuring such messages are
@@ -288,6 +292,7 @@ type Message struct {
 
 	MessageHash []byte // Hash of message. For rejects delivery, so optional like MessageID.
 	Flags
+	Keywords    []string `bstore:"index"` // Non-system or well-known $-flags. Only in "atom" syntax, stored in lower case.
 	Size        int64
 	TrainedJunk *bool  // If nil, no training done yet. Otherwise, true is trained as junk, false trained as nonjunk.
 	MsgPrefix   []byte // Typically holds received headers and/or header separator.
@@ -295,6 +300,7 @@ type Message struct {
 	// ParsedBuf message structure. Currently saved as JSON of message.Part because bstore
 	// cannot yet store recursive types. Created when first needed, and saved in the
 	// database.
+	// todo: once replaced with non-json storage, remove date fixup in ../message/part.go.
 	ParsedBuf []byte
 }
 
@@ -370,6 +376,17 @@ type Recipient struct {
 	Sent      time.Time      `bstore:"nonzero"`
 }
 
+// Outgoing is a message submitted for delivery from the queue. Used to enforce
+// maximum outgoing messages.
+type Outgoing struct {
+	ID        int64
+	Recipient string    `bstore:"nonzero,index"` // Canonical international address with utf8 domain.
+	Submitted time.Time `bstore:"nonzero,default now"`
+}
+
+// Types stored in DB.
+var DBTypes = []any{NextUIDValidity{}, Message{}, Recipient{}, Mailbox{}, Subscription{}, Outgoing{}, Password{}, Subjectpass{}}
+
 // Account holds the information about a user, includings mailboxes, messages, imap subscriptions.
 type Account struct {
 	Name   string     // Name, according to configuration.
@@ -386,13 +403,6 @@ type Account struct {
 	nused int // Reference count, while >0, this account is alive and shared.
 }
 
-func xcheckf(err error, format string, args ...any) {
-	if err != nil {
-		msg := fmt.Sprintf(format, args...)
-		panic(fmt.Errorf("%s: %w", msg, err))
-	}
-}
-
 // InitialUIDValidity returns a UIDValidity used for initializing an account.
 // It can be replaced during tests with a predictable value.
 var InitialUIDValidity = func() uint32 {
@@ -455,7 +465,7 @@ func openAccount(name string) (a *Account, rerr error) {
 		os.MkdirAll(dir, 0770)
 	}
 
-	db, err := bstore.Open(dbpath, &bstore.Options{Timeout: 5 * time.Second, Perm: 0660}, NextUIDValidity{}, Message{}, Recipient{}, Mailbox{}, Subscription{}, Password{}, Subjectpass{})
+	db, err := bstore.Open(context.TODO(), dbpath, &bstore.Options{Timeout: 5 * time.Second, Perm: 0660}, DBTypes...)
 	if err != nil {
 		return nil, err
 	}
@@ -484,7 +494,7 @@ func openAccount(name string) (a *Account, rerr error) {
 }
 
 func initAccount(db *bstore.DB) error {
-	return db.Write(func(tx *bstore.Tx) error {
+	return db.Write(context.TODO(), func(tx *bstore.Tx) error {
 		uidvalidity := InitialUIDValidity()
 
 		mailboxes := InitialMailboxes
@@ -568,7 +578,7 @@ func (a *Account) WithRLock(fn func()) {
 	fn()
 }
 
-// DeliverX delivers a mail message to the account.
+// DeliverMessage delivers a mail message to the account.
 //
 // If consumeFile is set, the original msgFile is moved/renamed or copied and
 // removed as part of delivery.
@@ -586,14 +596,16 @@ func (a *Account) WithRLock(fn func()) {
 // Must be called with account rlock or wlock.
 //
 // Caller must broadcast new message.
-func (a *Account) DeliverX(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os.File, consumeFile, isSent, sync, notrain bool) {
+func (a *Account) DeliverMessage(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os.File, consumeFile, isSent, sync, notrain bool) error {
 	mb := Mailbox{ID: m.MailboxID}
-	err := tx.Get(&mb)
-	xcheckf(err, "get mailbox")
+	if err := tx.Get(&mb); err != nil {
+		return fmt.Errorf("get mailbox: %w", err)
+	}
 	m.UID = mb.UIDNext
 	mb.UIDNext++
-	err = tx.Update(&mb)
-	xcheckf(err, "updating mailbox nextuid")
+	if err := tx.Update(&mb); err != nil {
+		return fmt.Errorf("updating mailbox nextuid: %w", err)
+	}
 
 	conf, _ := a.Conf()
 	m.JunkFlagsForMailbox(mb.Name, conf)
@@ -608,7 +620,9 @@ func (a *Account) DeliverX(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os
 		}
 		part = &p
 		buf, err := json.Marshal(part)
-		xcheckf(err, "marshal parsed message")
+		if err != nil {
+			return fmt.Errorf("marshal parsed message: %w", err)
+		}
 		m.ParsedBuf = buf
 	}
 
@@ -617,8 +631,9 @@ func (a *Account) DeliverX(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os
 		m.MailboxDestinedID = 0
 	}
 
-	err = tx.Insert(m)
-	xcheckf(err, "inserting message")
+	if err := tx.Insert(m); err != nil {
+		return fmt.Errorf("inserting message: %w", err)
+	}
 
 	if isSent {
 		// Attempt to parse the message for its To/Cc/Bcc headers, which we insert into Recipient.
@@ -658,8 +673,9 @@ func (a *Account) DeliverX(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os
 					OrgDomain: publicsuffix.Lookup(context.TODO(), d).Name(),
 					Sent:      sent,
 				}
-				err = tx.Insert(&mr)
-				xcheckf(err, "inserting sent message recipients")
+				if err := tx.Insert(&mr); err != nil {
+					return fmt.Errorf("inserting sent message recipients: %w", err)
+				}
 			}
 		}
 	}
@@ -670,30 +686,37 @@ func (a *Account) DeliverX(log *mlog.Log, tx *bstore.Tx, m *Message, msgFile *os
 
 	// Sync file data to disk.
 	if sync {
-		err = msgFile.Sync()
-		xcheckf(err, "fsync message file")
+		if err := msgFile.Sync(); err != nil {
+			return fmt.Errorf("fsync message file: %w", err)
+		}
 	}
 
 	if consumeFile {
-		err := os.Rename(msgFile.Name(), msgPath)
-		xcheckf(err, "moving msg file to destination directory")
+		if err := os.Rename(msgFile.Name(), msgPath); err != nil {
+			return fmt.Errorf("moving msg file to destination directory: %w", err)
+		}
 	} else if err := os.Link(msgFile.Name(), msgPath); err != nil {
 		// Assume file system does not support hardlinks. Copy it instead.
-		err := writeFile(msgPath, &moxio.AtReader{R: msgFile})
-		xcheckf(err, "copying message to new file")
+		if err := writeFile(msgPath, &moxio.AtReader{R: msgFile}); err != nil {
+			return fmt.Errorf("copying message to new file: %w", err)
+		}
 	}
 
 	if sync {
-		err = moxio.SyncDir(msgDir)
-		xcheckf(err, "sync directory")
+		if err := moxio.SyncDir(msgDir); err != nil {
+			return fmt.Errorf("sync directory: %w", err)
+		}
 	}
 
 	if !notrain && m.NeedsTraining() {
 		l := []Message{*m}
-		err = a.RetrainMessages(log, tx, l, false)
-		xcheckf(err, "training junkfilter")
+		if err := a.RetrainMessages(context.TODO(), log, tx, l, false); err != nil {
+			return fmt.Errorf("training junkfilter: %w", err)
+		}
 		*m = l[0]
 	}
+
+	return nil
 }
 
 // write contents of r to new file dst, for delivering a message.
@@ -726,7 +749,7 @@ func (a *Account) SetPassword(password string) error {
 		return fmt.Errorf("generating password hash: %w", err)
 	}
 
-	err = a.DB.Write(func(tx *bstore.Tx) error {
+	err = a.DB.Write(context.TODO(), func(tx *bstore.Tx) error {
 		if _, err := bstore.QueryTx[Password](tx).Delete(); err != nil {
 			return fmt.Errorf("deleting existing password: %v", err)
 		}
@@ -780,7 +803,7 @@ func (a *Account) SetPassword(password string) error {
 // Subjectpass returns the signing key for use with subjectpass for the given
 // email address with canonical localpart.
 func (a *Account) Subjectpass(email string) (key string, err error) {
-	return key, a.DB.Write(func(tx *bstore.Tx) error {
+	return key, a.DB.Write(context.TODO(), func(tx *bstore.Tx) error {
 		v := Subjectpass{Email: email}
 		err := tx.Get(&v)
 		if err == nil {
@@ -869,65 +892,50 @@ func (a *Account) MailboxEnsure(tx *bstore.Tx, name string, subscribe bool) (mb
 	return mb, changes, nil
 }
 
-// MailboxEnsureX calls MailboxEnsure, panicing with the error if it is not nil.
-func (a *Account) MailboxEnsureX(tx *bstore.Tx, name string, subscribe bool) (Mailbox, []Change) {
-	mb, changes, err := a.MailboxEnsure(tx, name, subscribe)
-	if err != nil {
-		panic(err)
-	}
-	return mb, changes
-}
-
-// Check if mailbox exists.
+// MailboxExists checks if mailbox exists.
 // Caller must hold account rlock.
-func (a *Account) MailboxExistsX(tx *bstore.Tx, name string) bool {
+func (a *Account) MailboxExists(tx *bstore.Tx, name string) (bool, error) {
 	q := bstore.QueryTx[Mailbox](tx)
 	q.FilterEqual("Name", name)
-	exists, err := q.Exists()
-	xcheckf(err, "checking existence")
-	return exists
+	return q.Exists()
 }
 
-// MailboxFindX finds a mailbox by name.
-func (a *Account) MailboxFindX(tx *bstore.Tx, name string) *Mailbox {
+// MailboxFind finds a mailbox by name, returning a nil mailbox and nil error if mailbox does not exist.
+func (a *Account) MailboxFind(tx *bstore.Tx, name string) (*Mailbox, error) {
 	q := bstore.QueryTx[Mailbox](tx)
 	q.FilterEqual("Name", name)
 	mb, err := q.Get()
 	if err == bstore.ErrAbsent {
-		return nil
+		return nil, nil
 	}
-	xcheckf(err, "lookup mailbox")
-	return &mb
+	if err != nil {
+		return nil, fmt.Errorf("looking up mailbox: %w", err)
+	}
+	return &mb, nil
 }
 
-// SubscriptionEnsureX ensures a subscription for name exists. The mailbox does not
+// SubscriptionEnsure ensures a subscription for name exists. The mailbox does not
 // have to exist. Any parents are not automatically subscribed.
-// Changes are broadcasted.
-func (a *Account) SubscriptionEnsureX(tx *bstore.Tx, name string) []Change {
-	err := tx.Get(&Subscription{name})
-	if err == nil {
-		return nil
+// Changes are returned and must be broadcasted by the caller.
+func (a *Account) SubscriptionEnsure(tx *bstore.Tx, name string) ([]Change, error) {
+	if err := tx.Get(&Subscription{name}); err == nil {
+		return nil, nil
 	}
 
-	err = tx.Insert(&Subscription{name})
-	xcheckf(err, "inserting subscription")
+	if err := tx.Insert(&Subscription{name}); err != nil {
+		return nil, fmt.Errorf("inserting subscription: %w", err)
+	}
 
 	q := bstore.QueryTx[Mailbox](tx)
 	q.FilterEqual("Name", name)
 	exists, err := q.Exists()
-	xcheckf(err, "looking up mailbox for subscription")
-	if exists {
-		return []Change{ChangeAddSubscription{name}}
+	if err != nil {
+		return nil, fmt.Errorf("looking up mailbox for subscription: %w", err)
 	}
-	return []Change{ChangeAddMailbox{Name: name, Flags: []string{`\Subscribed`, `\NonExistent`}}}
-}
-
-// List mailboxes. Only those that exist, so names with only a subscription are not returned.
-// Caller must have account rlock held.
-func (a *Account) MailboxesX(tx *bstore.Tx) []Mailbox {
-	l, err := bstore.QueryTx[Mailbox](tx).List()
-	xcheckf(err, "fetching mailboxes")
-	return l
+	if exists {
+		return []Change{ChangeAddSubscription{name}}, nil
+	}
+	return []Change{ChangeAddMailbox{Name: name, Flags: []string{`\Subscribed`, `\NonExistent`}}}, nil
 }
 
 // MessageRuleset returns the first ruleset (if any) that message the message
@@ -1038,21 +1046,23 @@ func (a *Account) Deliver(log *mlog.Log, dest config.Destination, m *Message, ms
 // Message delivery and possible mailbox creation are broadcasted.
 func (a *Account) DeliverMailbox(log *mlog.Log, mailbox string, m *Message, msgFile *os.File, consumeFile bool) error {
 	var changes []Change
-	err := extransact(a.DB, true, func(tx *bstore.Tx) error {
-		mb, chl := a.MailboxEnsureX(tx, mailbox, true)
+	err := a.DB.Write(context.TODO(), func(tx *bstore.Tx) error {
+		mb, chl, err := a.MailboxEnsure(tx, mailbox, true)
+		if err != nil {
+			return fmt.Errorf("ensuring mailbox: %w", err)
+		}
 		m.MailboxID = mb.ID
 		m.MailboxOrigID = mb.ID
 		changes = append(changes, chl...)
 
-		a.DeliverX(log, tx, m, msgFile, consumeFile, mb.Sent, true, false)
-		return nil
+		return a.DeliverMessage(log, tx, m, msgFile, consumeFile, mb.Sent, true, false)
 	})
 	// todo: if rename succeeded but transaction failed, we should remove the file.
 	if err != nil {
 		return err
 	}
 
-	changes = append(changes, ChangeAddUID{m.MailboxID, m.UID, m.Flags})
+	changes = append(changes, ChangeAddUID{m.MailboxID, m.UID, m.Flags, m.Keywords})
 	comm := RegisterComm(a)
 	defer comm.Unregister()
 	comm.Broadcast(changes)
@@ -1075,8 +1085,11 @@ func (a *Account) TidyRejectsMailbox(log *mlog.Log, rejectsMailbox string) (hasS
 		}
 	}()
 
-	err := extransact(a.DB, true, func(tx *bstore.Tx) error {
-		mb := a.MailboxFindX(tx, rejectsMailbox)
+	err := a.DB.Write(context.TODO(), func(tx *bstore.Tx) error {
+		mb, err := a.MailboxFind(tx, rejectsMailbox)
+		if err != nil {
+			return fmt.Errorf("finding mailbox: %w", err)
+		}
 		if mb == nil {
 			// No messages have been delivered yet.
 			hasSpace = true
@@ -1088,18 +1101,24 @@ func (a *Account) TidyRejectsMailbox(log *mlog.Log, rejectsMailbox string) (hasS
 		qdel := bstore.QueryTx[Message](tx)
 		qdel.FilterNonzero(Message{MailboxID: mb.ID})
 		qdel.FilterLess("Received", old)
-		var err error
 		remove, err = qdel.List()
-		xcheckf(err, "listing old messages")
+		if err != nil {
+			return fmt.Errorf("listing old messages: %w", err)
+		}
 
-		changes = a.xremoveMessages(log, tx, mb, remove)
+		changes, err = a.removeMessages(context.TODO(), log, tx, mb, remove)
+		if err != nil {
+			return fmt.Errorf("removing messages: %w", err)
+		}
 
 		// We allow up to n messages.
 		qcount := bstore.QueryTx[Message](tx)
 		qcount.FilterNonzero(Message{MailboxID: mb.ID})
 		qcount.Limit(1000)
 		n, err := qcount.Count()
-		xcheckf(err, "counting rejects")
+		if err != nil {
+			return fmt.Errorf("counting rejects: %w", err)
+		}
 		hasSpace = n < 1000
 
 		return nil
@@ -1116,9 +1135,9 @@ func (a *Account) TidyRejectsMailbox(log *mlog.Log, rejectsMailbox string) (hasS
 	return hasSpace, nil
 }
 
-func (a *Account) xremoveMessages(log *mlog.Log, tx *bstore.Tx, mb *Mailbox, l []Message) []Change {
+func (a *Account) removeMessages(ctx context.Context, log *mlog.Log, tx *bstore.Tx, mb *Mailbox, l []Message) ([]Change, error) {
 	if len(l) == 0 {
-		return nil
+		return nil, nil
 	}
 	ids := make([]int64, len(l))
 	anyids := make([]any, len(l))
@@ -1131,30 +1150,33 @@ func (a *Account) xremoveMessages(log *mlog.Log, tx *bstore.Tx, mb *Mailbox, l [
 	// from a Sent mailbox to the rejects mailbox...
 	qdmr := bstore.QueryTx[Recipient](tx)
 	qdmr.FilterEqual("MessageID", anyids...)
-	_, err := qdmr.Delete()
-	xcheckf(err, "deleting from message recipient")
+	if _, err := qdmr.Delete(); err != nil {
+		return nil, fmt.Errorf("deleting from message recipient: %w", err)
+	}
 
 	// Actually remove the messages.
 	qdm := bstore.QueryTx[Message](tx)
 	qdm.FilterIDs(ids)
 	var deleted []Message
 	qdm.Gather(&deleted)
-	_, err = qdm.Delete()
-	xcheckf(err, "deleting from messages")
+	if _, err := qdm.Delete(); err != nil {
+		return nil, fmt.Errorf("deleting from messages: %w", err)
+	}
 
 	// Mark as neutral and train so junk filter gets untrained with these (junk) messages.
 	for i := range deleted {
 		deleted[i].Junk = false
 		deleted[i].Notjunk = false
 	}
-	err = a.RetrainMessages(log, tx, deleted, true)
-	xcheckf(err, "training deleted messages")
+	if err := a.RetrainMessages(ctx, log, tx, deleted, true); err != nil {
+		return nil, fmt.Errorf("training deleted messages: %w", err)
+	}
 
 	changes := make([]Change, len(l))
 	for i, m := range l {
 		changes[i] = ChangeRemoveUIDs{mb.ID, []UID{m.UID}}
 	}
-	return changes
+	return changes, nil
 }
 
 // RejectsRemove removes a message from the rejects mailbox if present.
@@ -1172,19 +1194,26 @@ func (a *Account) RejectsRemove(log *mlog.Log, rejectsMailbox, messageID string)
 		}
 	}()
 
-	err := extransact(a.DB, true, func(tx *bstore.Tx) error {
-		mb := a.MailboxFindX(tx, rejectsMailbox)
+	err := a.DB.Write(context.TODO(), func(tx *bstore.Tx) error {
+		mb, err := a.MailboxFind(tx, rejectsMailbox)
+		if err != nil {
+			return fmt.Errorf("finding mailbox: %w", err)
+		}
 		if mb == nil {
 			return nil
 		}
 
 		q := bstore.QueryTx[Message](tx)
 		q.FilterNonzero(Message{MailboxID: mb.ID, MessageID: messageID})
-		var err error
 		remove, err = q.List()
-		xcheckf(err, "listing messages to remove")
+		if err != nil {
+			return fmt.Errorf("listing messages to remove: %w", err)
+		}
 
-		changes = a.xremoveMessages(log, tx, mb, remove)
+		changes, err = a.removeMessages(context.TODO(), log, tx, mb, remove)
+		if err != nil {
+			return fmt.Errorf("removing messages: %w", err)
+		}
 
 		return nil
 	})
@@ -1243,7 +1272,7 @@ func OpenEmailAuth(email string, password string) (acc *Account, rerr error) {
 		}
 	}()
 
-	pw, err := bstore.QueryDB[Password](acc.DB).Get()
+	pw, err := bstore.QueryDB[Password](context.TODO(), acc.DB).Get()
 	if err != nil {
 		if err == bstore.ErrAbsent {
 			return acc, ErrUnknownCredentials
@@ -1327,24 +1356,44 @@ func (f Flags) Set(mask, flags Flags) Flags {
 	return r
 }
 
-// FlagsQuerySet returns a map with the flags that are true in mask, with
-// values from flags.
-func FlagsQuerySet(mask, flags Flags) map[string]any {
-	r := map[string]any{}
-	set := func(f string, m, v bool) {
-		if m {
-			r[f] = v
+// RemoveKeywords removes keywords from l, modifying and returning it. Should only
+// be used with lower-case keywords, not with system flags like \Seen.
+func RemoveKeywords(l, remove []string) []string {
+	for _, k := range remove {
+		if i := slices.Index(l, k); i >= 0 {
+			copy(l[i:], l[i+1:])
+			l = l[:len(l)-1]
 		}
 	}
-	set("Seen", mask.Seen, flags.Seen)
-	set("Answered", mask.Answered, flags.Answered)
-	set("Flagged", mask.Flagged, flags.Flagged)
-	set("Forwarded", mask.Forwarded, flags.Forwarded)
-	set("Junk", mask.Junk, flags.Junk)
-	set("Notjunk", mask.Notjunk, flags.Notjunk)
-	set("Deleted", mask.Deleted, flags.Deleted)
-	set("Draft", mask.Draft, flags.Draft)
-	set("Phishing", mask.Phishing, flags.Phishing)
-	set("MDNSent", mask.MDNSent, flags.MDNSent)
-	return r
+	return l
+}
+
+// MergeKeywords adds keywords from add into l, updating and returning it along
+// with whether it added any keyword. Keywords are only added if they aren't
+// already present. Should only be used with lower-case keywords, not with system
+// flags like \Seen.
+func MergeKeywords(l, add []string) ([]string, bool) {
+	var changed bool
+	for _, k := range add {
+		if slices.Index(l, k) < 0 {
+			l = append(l, k)
+			changed = true
+		}
+	}
+	return l, changed
+}
+
+// ValidLowercaseKeyword returns whether s is a valid, lower-case, keyword.
+func ValidLowercaseKeyword(s string) bool {
+	for _, c := range s {
+		if c >= 'a' && c <= 'z' {
+			continue
+		}
+		// ../rfc/9051:6334
+		const atomspecials = `(){%*"\]`
+		if c <= ' ' || c > 0x7e || strings.ContainsRune(atomspecials, c) {
+			return false
+		}
+	}
+	return len(s) > 0
 }
